/* Implementation of GUI terminal on the Macintosh.
   Copyright (C) 1999, 2000 Free Software Foundation, Inc.
   Author: Andrew Choi <choi@cs.hku.hk>

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

// This is so we can get around the crappy mac os x header situation.
#define NON_OPAQUE

#ifdef TARGET_API_MAC_CARBON
  #undef DARWIN
  #undef macintosh
  #ifdef Z
    #undef Z
    #define __NEED_Z_DEFINE
  #endif
  #include <Carbon/Carbon.h>
  #define DARWIN
  #define macintosh
  #ifdef __NEED_Z_DEFINE
    #define Z (current_buffer->text->z)
  #endif
#else
  #include <Events.h>
  #include <Balloons.h>
  #include <Devices.h>
  #include <Fonts.h>
  #include <Gestalt.h>
  #include <Menus.h>
  #include <Processes.h>
  #include <Sound.h>
  #include <ToolUtils.h>
  #include <TextUtils.h>
  #include <Dialogs.h>
  #include <Script.h>
  #include <Scrap.h>
  #include <Types.h>
  #include <OSA.h>
  #include <AppleScript.h>
  #include <TextEncodingConverter.h>
  #include <Resources.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <ctype.h>

#if __MWERKS__
#include <unix.h>
#endif

#if __profile__
#include <profiler.h>
#endif

#include "config.h"
#include "lisp.h"
#include "termhooks.h"
#include "dispextern.h"
#include "frame.h"
#include "blockinput.h"
#include "charset.h"
#include "disptab.h"
#include "fontset.h"
#include "ccl.h"
#include "coding.h"
#include "keyboard.h"
#include "termchar.h"

#include "macterm.h"

#define mApple		128
#define   iAbout		1
#define mFile		129
#define   iQuit			1

#define rWindowResource	128

#define kAboutAlertID	128

#define kExtraStackAlloc (256 * 1024)
	
#ifndef TARGET_API_MAC_CARBON
QDGlobals qd;
#endif

Boolean	gDone = false;

const int kNumRows = 40;
const int kNumCols = 80;
const int kDefaultFontSize = 12;
const int kInsetPixels = 4;

const int kMinDocSize = 64;
const int kMaxDocSize = 32767;

/* updating_frame is set by update_begin in term.c to the frame being
   updated.  All termhook functions then implicitly modify this frame.
   It is set to zero by update_end.  Only mac_cursor_to,
   mac_write_glyphs, and mac_reassert_line_height may be called
   outside an update, in which case selected_frame refers to the frame
   that should be affected.  */
extern FRAME_PTR updating_frame;
extern FRAME_PTR selected_frame;

/* Nominal cursor position -- where to draw output during an update,
   these are different from the cursor-box position.  */
static int curs_x;
static int curs_y;

Lisp_Object Qinternal_border_width;
Lisp_Object Qforeground_color;
Lisp_Object Qbackground_color;
Lisp_Object Qvertical_scroll_bars;
Lisp_Object Qmouse_color;
Lisp_Object Qcursor_color;
Lisp_Object Qborder_color;
Lisp_Object Qmenu_bar_lines;
Lisp_Object Qscroll_bar_width;
Lisp_Object Qbuffer_predicate;
Lisp_Object Qicon_type;
Lisp_Object Qauto_raise;
Lisp_Object Qauto_lower;
Lisp_Object Qcursor_type;
Lisp_Object Qbox;
Lisp_Object Qnone;



extern Lisp_Object Qtitle, Qonly, Qminibuffer;

extern void InitEmacsPasswdDir ();

/* true if using command key as meta key */
Lisp_Object Vmac_command_key_is_meta;

/* convert input from Mac keyboard (assumed to be in Mac Roman coding)
   to this text encoding */
int mac_keyboard_text_encoding;
int current_mac_keyboard_text_encoding = kTextEncodingMacRoman;

Lisp_Object Qleft;
Lisp_Object Qtop;

#define RED_FROM_ULONG(color) (color >> 16)
#define GREEN_FROM_ULONG(color) ((color >> 8) & 0xff)
#define BLUE_FROM_ULONG(color) (color & 0xff)

/* Return the upper/left pixel position of the character cell on frame
   F at ROW/COL.  */
#define CHAR_TO_PIXEL_ROW(f, row) \
  (kInsetPixels + (row) * (f)->output_data.mac->mLineHeight)
#define CHAR_TO_PIXEL_COL(f, col) \
  (kInsetPixels + (col) * FONT_WIDTH ((f)->output_data.mac->font))

/* Return the row/column (zero-based) of the character cell containing
   the pixel on FRAME at ROW/COL.  */
#define PIXEL_TO_CHAR_ROW(f, row) \
  (((row) - kInsetPixels) \
   / (f)->output_data.mac->mLineHeight)
#define PIXEL_TO_CHAR_COL(f, col) \
  (((col) - kInsetPixels) \
   / FONT_WIDTH ((f)->output_data.mac->font))

#define FONT_BASE(f)	((f)->mAscent)

/* MPW MrC treats (~0) as 1 with -opt local option */
#define FACE_DEFAULT 0xffffffff

/* sleep time for WaitNextEvent */
#define WNE_SLEEP_AT_SUSPEND 10
#define WNE_SLEEP_AT_RESUME  1

/* true between calls to mac_update_begin and mac_update_end */
static int gUpdateInProgress = 0;
/* true when cannot handle any Mac OS events */
static int gHandlingWindowUpdate = 0;

/* the flag gFlagAppSuspend is used both for determining the sleep
   time to be passed to WaitNextEvent and whether the cursor should be
   drawn when updating the display.  The cursor is turned off when
   Emacs is suspended.  Redrawing it is unnecessary and what needs to
   be done depends on whether the cursor lies inside or outside the
   redraw region.  So we might as well skip drawing it when Emacs is
   suspended.  */
static Boolean gFlagAppSuspend = false;
static long gAppSleepTime = WNE_SLEEP_AT_RESUME;

void doInitManagers (void);
void doGetMenus (void);
void doEvents (EventRecord *);
void doOSEvent (EventRecord *);
void doAdjustMenus (void);
void doMenuChoice (SInt32);
void doAppleMenu (SInt16);
void doFileMenu (SInt16);
static void doGrowWindow (WindowPtr, EventRecord *);
static void doZoomWindow (WindowPtr, int);
static void resizeHelper (WindowPtr, int, int);

static void mac_windows_initialize();
void mac_cursor_to (int X, int Y);
void mac_clear_to_end ();
void mac_clear_frame ();
void mac_clear_end_of_line (int);
void mac_ins_del_lines (int, int);
void mac_reassert_line_highlight (int, int);
void mac_change_line_highlight (int, int, int);
static void mac_write_glyphs (GLYPH *, int);
void mac_ring_bell ();
void mac_set_terminal_modes ();
void x_set_frame_parameters (FRAME_PTR f, Lisp_Object alist);
void mac_set_frame_parameters (FRAME_PTR f, Lisp_Object list);
void x_make_frame_visible (struct frame *f);
void mac_reset_terminal_modes ();
void mac_update_begin (struct frame *);
void mac_update_end (struct frame *);
void mac_set_terminal_window (int);
void mac_frame_raise_lower (struct frame *, int);
int mac_read_socket (int, struct input_event *, int, int);
void mac_mouse_position (struct frame **, int, Lisp_Object *,
			 enum scroll_bar_part *,
			 Lisp_Object *x,
			 Lisp_Object *y,
			 unsigned long *time);
mac_output *NewMacWindow ();
void DisposeMacWindow (mac_output *);

Lisp_Object Vx_resource_name;

static void CursorOn (struct mac_output *mwp);
static void CursorOff (struct mac_output *mwp);

static void InitializeAppleScript();
static void TerminateAppleScript();

/* Apple Events */
static void InitRequiredAppleEvents(void);
static pascal OSErr doAEOpenApplication(const AppleEvent *pAE, AppleEvent *pReply, long pRefcon);
static pascal OSErr doAEPrintDocuments(const AppleEvent *pAE, AppleEvent *pReply, long pRefcon);
static pascal OSErr doAEOpenDocuments(AppleEvent *message, AppleEvent *, long);
static pascal OSErr doAEQuitApplication(AppleEvent *message, AppleEvent *, long);

Lisp_Object gDragAndDropFileList;

#define ArgvStringListID 129

/* the global font table */
struct font_info *font_table;
int font_table_size = 0;
int n_fonts = 0;

Lisp_Object Qface_set_after_frame_default;

void mac_set_frame_fontset (struct frame *f, Lisp_Object arg, Lisp_Object oldVal);

char **gFontNameTable = NULL;
int gFontNameTableSize = 0;
int gFontNameCount = 0;

#if __profile__
void
ProfilerExitProc ()
{
  ProfilerDump ("\pEmacs.prof");
  ProfilerTerm ();
}
#endif

/* These few functions implement Emacs as a normal Mac application
   (almost): set up the the heap and the Toolbox, handle necessary
   system events plus a few simple menu events.  They also set up
   Emacs's access to functions defined in the rest of this file.
   Emacs uses function hooks to perform all its terminal I/O.  A
   complete list of these functions appear in termhooks.h.  For what
   they do, read the comments there and see also w32term.c and
   xterm.c.  What's noticeably missing here is the event loop, which
   is normally present in most Mac application.  After performing the
   necessary Mac initializations, main passes off control to
   emacs_main (corresponding to main in emacs.c).  Emacs_main calls
   mac_read_socket (defined further below) to read input.  This is
   where WaitNextEvent is called to process Mac events.  This is also
   where CheckAlarm in sysdep.c is called to simulate alarm signals.
   This makes the cursor jump back to its correct position after
   briefly jumping to that of the matching parenthesis, print useful
   hints and prompts in the minibuffer after the user stops typing for
   a wait, etc.  */

#undef main
#if !defined(TARGET_API_MAC_CARBON)
int 
main (int c, char **v, char **e)
{
    printf("Alive!!!\n");
    mac_windows_initialize();
    {
	char **argv;
	int argc = 0;

	/* set up argv array from STR# resource */
	GetStringList (&argv, ArgvStringListID);
	while (argv[argc])
	    argc++;

	/* 3rd param "envp" never used in emacs_main */
	(void) emacs_main (argc, argv, 0);
    }

    /* Never reached - real exit in Fkill_emacs */
}
#endif

static void mac_windows_initialize()
{
#if __profile__  /* is the profiler on? */
  if (ProfilerInit(collectDetailed, bestTimeBase, 5000, 200))
    exit(1);
#endif

#if __MWERKS__
  /* set creator and type for files created by MSL */
  _fcreator = 'EMAx';
  _ftype = 'TEXT';
#endif

  doInitManagers ();
	
  doGetMenus ();
	
#if !defined(TARGET_API_MAC_CARBON)
  InitEmacsPasswdDir ();
  InitEnviron ();
#endif

  InitializeAppleScript();

  /* Register basic apple event handlers */
  InitRequiredAppleEvents();

  cursor_to_hook = mac_cursor_to;
  raw_cursor_to_hook = mac_cursor_to;
  clear_to_end_hook = mac_clear_to_end;
  clear_frame_hook = mac_clear_frame;
  clear_end_of_line_hook = mac_clear_end_of_line;
  ins_del_lines_hook = mac_ins_del_lines;
  change_line_highlight_hook = mac_change_line_highlight;
  reassert_line_highlight_hook = mac_reassert_line_highlight;
  write_glyphs_hook = mac_write_glyphs;
  ring_bell_hook = mac_ring_bell;
  set_terminal_modes_hook = mac_set_terminal_modes;
  reset_terminal_modes_hook = mac_reset_terminal_modes;
  update_begin_hook = mac_update_begin;
  update_end_hook = mac_update_end;
  set_terminal_window_hook = mac_set_terminal_window;
  frame_raise_lower_hook = mac_frame_raise_lower;
  read_socket_hook = mac_read_socket;
  mouse_position_hook = mac_mouse_position;
	
  baud_rate = 4*38400;

  /* free up AppleScript resources on exit */
  atexit (TerminateAppleScript);

#if __profile__  /* is the profiler on? */
  atexit (ProfilerExitProc);
#endif
}

void
doInitManagers (void)
{
#if !defined(TARGET_API_MAC_CARBON)
  InitGraf (&qd.thePort);
  InitFonts ();
  FlushEvents (everyEvent,0);
  InitWindows ();
  InitMenus ();
  TEInit ();
  InitDialogs (NULL);
  InitCursor ();	

  /* set up some extra stack space for use by emacs */
  SetApplLimit ((Ptr) ((long) GetApplLimit () - kExtraStackAlloc));

  /* MaxApplZone must be called for AppleScript to execute more
     complicated scripts */
  MaxApplZone ();
  MoreMasters ();
#endif
}

void
doGetMenus (void)
{
#if !defined(TARGET_API_MAC_CARBON)
  Handle menubarHdl;
  MenuHandle menuHdl;
  OSErr osErr;
	
  menubarHdl = GetNewMBar (128);
  if(menubarHdl == NULL)
    ExitToShell ();
  SetMenuBar (menubarHdl);
  DrawMenuBar ();

  menuHdl = GetMenuHandle (mApple);
  if(menuHdl != NULL)
    AppendResMenu (menuHdl,'DRVR');
  else
    ExitToShell ();
#endif
}

static void
doWindowUpdate (WindowPtr win)
{
  struct mac_output *mwp;
  mwp = (mac_output *) GetWRefCon (win);

  BeginUpdate (win);

  /* Set gHandlingWindowUpdate flag to temporarily turn off polling of
     events (checked in mac_read_socket).  Otherwise a new update
     event preempts an ongoing screen redraw, leaving blank spots on
     screen.  Note that BLOCK_INPUT and stop_polling have no effect
     under the Mac OS configurations.  */
  gHandlingWindowUpdate = 1;
  
  SET_FRAME_GARBAGED (mwp->mFP); 
  redraw_garbaged_frames (); 
  frame_garbaged = 0; 
  redisplay (); 

  gHandlingWindowUpdate = 0;

  EndUpdate (win);
}

static void
doWindowActivate (WindowPtr win)
{
  CursorOn ((mac_output *) GetWRefCon (win));
}

static void
doWindowDeactivate(WindowPtr win)
{
  CursorOff ((mac_output *) GetWRefCon (win));
}

static void
doAppSuspend ()
{
  gFlagAppSuspend = true;
  gAppSleepTime = WNE_SLEEP_AT_SUSPEND;
  CursorOff ((mac_output *) GetWRefCon (FrontWindow ()));
}

static void
doAppResume ()
{
  gFlagAppSuspend = false;
  gAppSleepTime = WNE_SLEEP_AT_RESUME;
  CursorOn ((mac_output *) GetWRefCon (FrontWindow ()));
}

void
doEvents (EventRecord *eventStrucPtr)
{
  SInt8	charCode;

  switch (eventStrucPtr->what)
    {
    case updateEvt:
      doWindowUpdate ((WindowPtr) eventStrucPtr->message);
      break;

    case osEvt:
      doOSEvent (eventStrucPtr);
      break;

    case activateEvt:
      if ((eventStrucPtr->modifiers & activeFlag) != 0)
	doWindowActivate ((WindowPtr) eventStrucPtr->message);
      else
	doWindowDeactivate ((WindowPtr) eventStrucPtr->message);
      break;
    }
}

void
doOSEvent (EventRecord *eventStrucPtr)
{
  switch((eventStrucPtr->message >> 24) & 0x000000FF)
    {
    case suspendResumeMessage:
      if((eventStrucPtr->message & resumeFlag) == 1)
	{
	  doAppResume();  /* put future resume code here */
	}
      else
	{
	  doAppSuspend(); /* put future suspend code here */
	}
      break;
				
    case mouseMovedMessage:
      /* put future mouse moved code here */
      break;
    }
}

void
doAdjustMenus (void)
{
  /* Adjust menus here. */
}

void
doMenuChoice (SInt32 menuChoice)
{
  SInt16 menuID, menuItem;
		
  menuID = HiWord (menuChoice);
  menuItem = LoWord (menuChoice);

  if (menuID == 0)
    return;

  switch (menuID)
    {
    case mApple:
      doAppleMenu (menuItem);
      break;

    case mFile:
      doFileMenu (menuItem);
      break;
    }
  
  HiliteMenu (0);
}

void
doAppleMenu (SInt16 menuItem)
{
#if !defined(TARGET_API_MAC_CARBON)
  Str255 itemName;
  SInt16 daDriverRefNum;
	
  if (menuItem == iAbout)
    NoteAlert (kAboutAlertID, NULL);
  else
    {
      GetMenuItemText (GetMenuHandle (mApple), menuItem, itemName);
      daDriverRefNum = OpenDeskAcc (itemName);
    }
#endif
}

void
doFileMenu (SInt16 menuItem)
{
  if (menuItem  == iQuit)
    gDone = true;

  return;

  /* Fkill_emacs (make_number (1)); */
  
  /* Should bind this to save-buffers-kill-emacs using the menu
     subsystem later.  Not as bad as it looks because Fkill_emacs at
     least does an autosave.  */
}

void MacTweak()
{
  WindowPtr window;
  Rect r;
  MacSetRect(&r,20,50,30+20,30+50);
  //  if (!(window = NewCWindow(NULL,&r,"\005Emacs\0",true/*visible*/,documentProc, (WindowPtr) -1,true/*goawayflag*/,0/*refcon*/)))
  window = NewCWindow(NULL,&r,"\005Emacs\0",true/*visible*/,kWindowGrowDocumentProc, (WindowPtr) -1,true/*goawayflag*/,0/*refcon*/);
  //    while (!Button()) {}
  //    while (Button()) {}
}

/* Here comes the set of functions Emacs will call to perform all its
   terminal I/O.  For each window on the screen, which Emacs calls a
   "frame", Emacs maintains a window system independent struct frame
   (defined in frame.h) to store its characteristics many of which can
   be accessed by elisp code.  Among the fields of the struct is a C
   union output_data which stores a pointer to a struct mac_output,
   which we define in macterm.h for storing Mac dependent window
   characteristics (such as font number, font size).  This struct also
   contains a QuickDraw WindowPtr, which serves as a handle to the
   corresponding MacOS window.  These three structs are doubly linked
   so given a pointer to one, the other two can also be accessed.
   Perhaps worth mentioning is that the MacOS window uses the RefCon
   field to store a pointer back to its mac_output struct.  */

/* Initialize the struct pointed to by MW to represent a new COLS x
   ROWS Macintosh window, using font with name FONTNAME and size
   FONTSIZE.  */
mac_output *
NewMacWindow (FRAME_PTR fp)
{
  mac_output *mwp;
  FontInfo theFontInfo;
  Rect r;
  char **systemName;
  int i;

  if (!(mwp = (mac_output *) xmalloc (sizeof (struct mac_output))))
    {
      SysBeep (10);
      ExitToShell ();
    }

#if defined(TARGET_API_MAC_CARBON)
  MacSetRect(&r,20,50,640+20,480+50);
  //  if (!(mwp->mWP = NewCWindow(NULL,&r,"\005Emacs\0",true/*visible*/,documentProc, (WindowPtr) -1,true/*goawayflag*/,0/*refcon*/)))
  if (!(mwp->mWP = NewCWindow(NULL,&r,"\005Emacs\0",true/*visible*/,kWindowGrowDocumentProc, (WindowPtr) -1,true/*goawayflag*/,0/*refcon*/)))
#else
  if (!(mwp->mWP = GetNewCWindow (rWindowResource, NULL, (WindowPtr) -1)))
#endif
    {
      SysBeep (10);
      ExitToShell ();
    }

  SetWRefCon (mwp->mWP, (long) mwp);  /* so that update events can find this mac_output struct */
  mwp->mFP = fp;  /* point back to emacs frame */

  SetPort (GetWindowPort(mwp->mWP));

  mwp->mNumCols = kNumCols;	
  mwp->mNumRows = kNumRows;

  /* mwp->font used to be set to a (XFontStruct*) (~0) to prevent
     slots in COMPUTED_FACES that represent the default face from
     being overwritten (zero denotes free slots).  Without this, in
     fontlock mode, default text (normally black on white) will be
     displayed in the first colored face loaded (normally red for
     comments).  This bug fix is due to Sudhir Shenoy. */
  mwp->font = (MacFontStruct *) xmalloc (sizeof (struct MacFontStruct));
  mwp->fontset = -1;
  
  /* Use preferred font and size for system script */
  mwp->font->mFontNum = HiWord (GetScriptVariable (smSystemScript, smScriptMonoFondSize));
  mwp->font->mFontSize = LoWord (GetScriptVariable (smSystemScript, smScriptMonoFondSize));
  mwp->font->mFontFace = normal;
  TextFont (mwp->font->mFontNum);
  TextSize (mwp->font->mFontSize);
  TextFace (mwp->font->mFontFace);
  mwp->font->mScriptCode = FontToScript (mwp->font->mFontNum);
  mwp->font->mTwoByte = mwp->font->mScriptCode == smJapanese ||
                        mwp->font->mScriptCode == smTradChinese ||
                        mwp->font->mScriptCode == smSimpChinese ||
                        mwp->font->mScriptCode == smKorean;

  GetFontInfo (&theFontInfo);
  mwp->font->mHeight = theFontInfo.ascent + theFontInfo.descent + theFontInfo.leading;
  /* mwp->mCharWidth = theFontInfo.widMax; Incorrectly returns 15 for
     12-point Monaco!  */
  mwp->font->mWidth = CharWidth ('m');

  mwp->font->mAscent = theFontInfo.ascent;
  mwp->font->mDescent = theFontInfo.descent;
  mwp->font->mLeading = theFontInfo.leading;

  mwp->mLineHeight = mwp->font->mHeight;
  mwp->mCharWidth = mwp->font->mWidth;
  SizeWindow (mwp->mWP, mwp->mCharWidth * kNumCols + 2 * kInsetPixels, mwp->mLineHeight * kNumRows + 2 * kInsetPixels, false);

#if defined(TARGET_API_MAC_CARBON)
  {
      char name[MAXHOSTNAMELEN] = "",*sysName;
      gethostname(name,sizeof(name));
      sysName = strdup(name);
      name[0] = snprintf(&name[1],sizeof(name)-1,"Emacs on %s",sysName);
      SetWTitle(mwp->mWP, name);
      free(sysName);
  }
#else
  /* Ugly magic Mac thing; see IM - Resource Manager Ref.  */
  systemName = (char**)GetString (-16413);
  if (systemName)
    SetWTitle (mwp->mWP, *systemName);
#endif
  ShowWindow (mwp->mWP);
	
  mwp->mHomeX = kInsetPixels;
  mwp->mHomeY = kInsetPixels + theFontInfo.ascent + theFontInfo.leading / 2;
		
  /* Put the cursor at initial position.  */
  r.left = kInsetPixels;
  r.right = r.left + mwp->mCharWidth;
  r.top = kInsetPixels;
  r.bottom = r.top + mwp->mLineHeight;

  InvertRect (&r);
  
  mwp->mTermWinSize = kNumRows - 1;
	
  return mwp;
}

void
DisposeMacWindow (mac_output *mwp)
{
  //  CloseWindow (mwp->mWP);
  //  DisposePtr ((Ptr)(mwp->mWP));
  DisposeWindow(mwp->mWP);
  free (mwp);
}

/* Types we might convert a resource string into.  */
enum resource_types
  {
    number, boolean, string, symbol
  };

/* Return the value of parameter PARAM.

   First search ALIST, then Vdefault_frame_alist, then the X defaults
   database, using ATTRIBUTE as the attribute name and CLASS as its class.

   Convert the resource to the type specified by desired_type.

   If no default is specified, return Qunbound.  If you call
   x_get_arg, make sure you deal with Qunbound in a reasonable way,
   and don't let it get stored in any Lisp-visible variables!  */

static Lisp_Object
x_get_arg (alist, param, attribute, class, type)
     Lisp_Object alist, param;
     char *attribute;
     char *class;
     enum resource_types type;
{
  register Lisp_Object tem;

  tem = Fassq (param, alist);
  if (EQ (tem, Qnil))
    tem = Fassq (param, Vdefault_frame_alist);
  if (EQ (tem, Qnil))
    {

      if (attribute)
	{
	  tem = Fx_get_resource (build_string (attribute),
				 build_string (class),
				 Qnil, Qnil);

	  if (NILP (tem))
	    return Qunbound;

	  switch (type)
	    {
	    case number:
	      return make_number (atoi (XSTRING (tem)->data));

	    case boolean:
	      tem = Fdowncase (tem);
	      if (!strcmp (XSTRING (tem)->data, "on")
		  || !strcmp (XSTRING (tem)->data, "true"))
		return Qt;
	      else 
		return Qnil;

	    case string:
	      return tem;

	    case symbol:
	      /* As a special case, we map the values `true' and `on'
		 to Qt, and `false' and `off' to Qnil.  */
	      {
		Lisp_Object lower;
		lower = Fdowncase (tem);
		if (!strcmp (XSTRING (lower)->data, "on")
		    || !strcmp (XSTRING (lower)->data, "true"))
		  return Qt;
		else if (!strcmp (XSTRING (lower)->data, "off")
		      || !strcmp (XSTRING (lower)->data, "false"))
		  return Qnil;
		else
		  return Fintern (tem, Qnil);
	      }

	    default:
	      abort ();
	    }
	}
      else
	return Qunbound;
    }
  return Fcdr (tem);
}

/* Record in frame F the specified or default value according to ALIST
   of the parameter named PARAM (a Lisp symbol).
   If no value is specified for PARAM, look for an X default for XPROP
   on the frame named NAME.
   If that is not found either, use the value DEFLT.  */

static Lisp_Object
x_default_parameter (f, alist, prop, deflt, xprop, xclass, type)
     struct frame *f;
     Lisp_Object alist;
     Lisp_Object prop;
     Lisp_Object deflt;
     char *xprop;
     char *xclass;
     enum resource_types type;
{
  Lisp_Object tem;

  tem = x_get_arg (alist, prop, xprop, xclass, type);
  if (EQ (tem, Qunbound))
    tem = deflt;
  x_set_frame_parameters (f, Fcons (Fcons (prop, tem), Qnil));
  return tem;
}

DEFUN ("x-create-frame", Fx_create_frame, Sx_create_frame,
       1, 1, 0,
  "Make a new window, which is called a \"frame\" in Emacs terms.\n\
Returns an Emacs frame object.\n\
ALIST is an alist of frame parameters.\n\
If the parameters specify that the frame should not have a minibuffer,\n\
and do not specify a specific minibuffer window to use,\n\
then `default-minibuffer-frame' must be a frame whose minibuffer can\n\
be shared by the new frame.\n\
\n\
This function is an internal primitive--use `make-frame' instead.")
  (parms)
     Lisp_Object parms;
{
  int count = specpdl_ptr - specpdl;
  struct frame *f;
  struct gcpro gcpro1,gcpro2;
  Lisp_Object frame, tem;
  char name[20];
  static mac_initialized = 0;
  KBOARD *kb;
  int minibuffer_only = 0;
  
//  fprintf(stderr,"x-create-frame\n");

//  if (selected_frame->output_method != output_mac)
//    error ("Not running on a Macintosh screen; cannot make a new Macintosh frame");

#ifdef MULTI_KBOARD
  if (!initial_kboard)
    {
      initial_kboard = (KBOARD *) xmalloc (sizeof (KBOARD));
      init_kboard (initial_kboard);
      initial_kboard->next_kboard = all_kboards;
      all_kboards = initial_kboard;
    }
  kb = initial_kboard;
#else
  kb = &the_only_kboard;
#endif

  if (!mac_initialized) {
      mac_windows_initialize();
      mac_initialized = 1;
  }
  
  /* The first call must initialize Vframe_list.  */
  if (! (NILP (Vframe_list) || CONSP (Vframe_list)))
    Vframe_list = Qnil;

#if 0
  name = x_get_arg (parms, Qname, "name", "Name", string);
  if (!STRINGP (name)
      && ! EQ (name, Qunbound)
      && ! NILP (name))
    error ("Invalid frame name--not a string or nil");
  
  if (STRINGP (name))
    Vx_resource_name = name;
#endif

  frame = Qnil;
  GCPRO2(parms,frame);
  tem = x_get_arg (parms, Qminibuffer, "minibuffer", "Minibuffer", symbol);
  if (EQ (tem, Qnone) || NILP (tem))
      f = make_frame_without_minibuffer (Qnil, kb, Qnil/*display*/);
  else if (EQ (tem, Qonly))
  {
      f = make_minibuffer_frame ();
      minibuffer_only = 1;
  }
  else if (WINDOWP (tem))
      f = make_frame_without_minibuffer (tem, kb, Qnil/*display*/);
  else
      f = make_frame (1);

  XSETFRAME (frame, f);
  Vframe_list = Fcons (frame, Vframe_list);

//  terminal_frame_count++;
//  sprintf (name, "F%d", terminal_frame_count);
//  f->name = build_string (name);
  f->name = build_string ("Hola!");

  f->visible = 1;		/* FRAME_SET_VISIBLE wd set frame_garbaged. */
  f->async_visible = 1;		/* Don't let visible be cleared later. */

  f->output_data.mac = NewMacWindow(f);
  f->output_data.mac->background_pixel = 0xffffff;
  f->output_data.mac->foreground_pixel = 0;
  f->output_data.mac->cursor_pixel = f->output_data.mac->foreground_pixel;
  f->output_data.mac->n_param_faces = 0;
  f->output_data.mac->n_computed_faces = 0;
  f->output_data.mac->size_computed_faces = 0;  
  f->output_method = output_mac;
  f->auto_raise = 1;
  f->auto_lower = 1;
  init_frame_faces (f);

  /* This stuff has to come after we make the frame because it relies
     on the structure and window existing. This stuff will crash if
     you move the NewMacWindow() call after this section (ie, if you
     try to make it look like x-create-display in xfns.c. */
  
  /* Default internalBorderWidth to 0 on Windows to match other programs.  */
  x_default_parameter (f, parms, Qinternal_border_width, make_number (0),
		       "internalBorderWidth", "BorderWidth", number);
  x_default_parameter (f, parms, Qvertical_scroll_bars, Qt,
		       "verticalScrollBars", "ScrollBars", boolean);

  /* Also do the stuff which must be set before the window exists.  */
  x_default_parameter (f, parms, Qforeground_color, build_string ("black"),
		       "foreground", "Foreground", string);
  x_default_parameter (f, parms, Qbackground_color, build_string ("white"),
		       "background", "Background", string);
  x_default_parameter (f, parms, Qmouse_color, build_string ("black"),
		       "pointerColor", "Foreground", string);
  x_default_parameter (f, parms, Qcursor_color, build_string ("black"),
		       "cursorColor", "Foreground", string);
  x_default_parameter (f, parms, Qborder_color, build_string ("black"),
		       "borderColor", "BorderColor", string);

  x_default_parameter (f, parms, Qmenu_bar_lines, make_number (1),
		       "menuBar", "MenuBar", number);
  x_default_parameter (f, parms, Qscroll_bar_width, Qnil,
		       "scrollBarWidth", "ScrollBarWidth", number);
  x_default_parameter (f, parms, Qbuffer_predicate, Qnil,
		       "bufferPredicate", "BufferPredicate", symbol);
  x_default_parameter (f, parms, Qtitle, Qnil,
		       "title", "Title", string);

  /* We need to do this after creating the window, so that the
     icon-creation functions can say whose icon they're describing.  */
  x_default_parameter (f, parms, Qicon_type, Qnil,
		       "bitmapIcon", "BitmapIcon", symbol);

  x_default_parameter (f, parms, Qauto_raise, Qnil,
		       "autoRaise", "AutoRaiseLower", boolean);
  x_default_parameter (f, parms, Qauto_lower, Qnil,
		       "autoLower", "AutoRaiseLower", boolean);
  x_default_parameter (f, parms, Qcursor_type, Qbox,
		       "cursorType", "CursorType", symbol);

#ifdef MULTI_KBOARD
  FRAME_KBOARD (f) = kb;
#endif

#if !defined(macintosh)  /* ???? */
  f->output_data.nothing = 1;	/* Nonzero means frame isn't deleted.  */
#endif

//  change_frame_size (f, FRAME_HEIGHT (selected_frame),
//		     FRAME_WIDTH (selected_frame), 0, 0);
  change_frame_size (f, f->output_data.mac->mNumRows,
                     f->output_data.mac->mNumCols, 0, 0);
  remake_frame_glyphs (f);
  calculate_costs (f);
//  XSETFRAME (frame, f);
  Fmodify_frame_parameters (frame, Vdefault_frame_alist);
  Fmodify_frame_parameters (frame, parms);
  f->face_alist = selected_frame->face_alist;

  UNGCPRO;

  /* It is now ok to make the frame official
     even if we get an error below.
     And the frame needs to be on Vframe_list
     or making it visible won't work.  */
  Vframe_list = Fcons (frame, Vframe_list);

  /* let it appear */
  x_make_frame_visible (f);

  return unbind_to (count, frame);
}

/* Set foreground and background color for subsequent QuickDraw
   drawing, assuming that the graphic port has already been set.  */
static void
SetFrameFgBgColors (unsigned long fgc, unsigned long bgc)
{
  RGBColor fgColor, bgColor;
						
  fgColor.red = RED_FROM_ULONG (fgc) * 256;
  fgColor.green = GREEN_FROM_ULONG (fgc) * 256;
  fgColor.blue = BLUE_FROM_ULONG (fgc) * 256;
			
  RGBForeColor (&fgColor);
			
  bgColor.red = RED_FROM_ULONG (bgc) * 256;
  bgColor.green = GREEN_FROM_ULONG (bgc) * 256;
  bgColor.blue = BLUE_FROM_ULONG (bgc) * 256;
			
  RGBBackColor (&bgColor);
}

/* Make sure that Vx_resource_name is set to a reasonable value.
   Fix it up, or set it to `emacs' if it is too hopeless.  */

static void
validate_x_resource_name ()
{
  int len;
  /* Number of valid characters in the resource name.  */
  int good_count = 0;
  /* Number of invalid characters in the resource name.  */
  int bad_count = 0;
  Lisp_Object new;
  int i;

  if (STRINGP (Vx_resource_name))
    {
      unsigned char *p = XSTRING (Vx_resource_name)->data;
      int i;

      len = XSTRING (Vx_resource_name)->size;

      /* Only letters, digits, - and _ are valid in resource names.
	 Count the valid characters and count the invalid ones.  */
      for (i = 0; i < len; i++)
	{
	  int c = p[i];
	  if (! ((c >= 'a' && c <= 'z')
		 || (c >= 'A' && c <= 'Z')
		 || (c >= '0' && c <= '9')
		 || c == '-' || c == '_'))
	    bad_count++;
	  else
	    good_count++;
	}
    }
  else
    /* Not a string => completely invalid.  */
    bad_count = 5, good_count = 0;

  /* If name is valid already, return.  */
  if (bad_count == 0)
    return;

  /* If name is entirely invalid, or nearly so, use `emacs'.  */
  if (good_count == 0
      || (good_count == 1 && bad_count > 0))
    {
      Vx_resource_name = build_string ("emacs");
      return;
    }

  /* Name is partly valid.  Copy it and replace the invalid characters
     with underscores.  */

  Vx_resource_name = new = Fcopy_sequence (Vx_resource_name);

  for (i = 0; i < len; i++)
    {
      int c = XSTRING (new)->data[i];
      if (! ((c >= 'a' && c <= 'z')
	     || (c >= 'A' && c <= 'Z')
	     || (c >= '0' && c <= '9')
	     || c == '-' || c == '_'))
	XSTRING (new)->data[i] = '_';
    }
}

#define EMACS_CLASS "Emacs"

/* Retrieve the string resource specified by NAME with CLASS from
   database RDB. */

char *
x_get_string_resource (rdb, name, class)
void *rdb;//     XrmDatabase rdb;
     char *name, *class;
{
    return NULL;
    if (strstr(name,"background-color"))
        return "white";
    if (strstr(name,"foreground-color"))
        return "black";
    if (strstr(name,"border-color"))
        return "grey";
    if (strstr(name,"mouse-color"))
        return "pink";
    if (strcmp(name,"cursor-color"))
        return "CadetBlue1";

    return NULL;//"no";
}

DEFUN ("x-get-resource", Fx_get_resource, Sx_get_resource, 2, 4, 0,
  "Return the value of ATTRIBUTE, of class CLASS, from the X defaults database.\n\
This uses `INSTANCE.ATTRIBUTE' as the key and `Emacs.CLASS' as the\n\
class, where INSTANCE is the name under which Emacs was invoked, or\n\
the name specified by the `-name' or `-rn' command-line arguments.\n\
\n\
The optional arguments COMPONENT and SUBCLASS add to the key and the\n\
class, respectively.  You must specify both of them or neither.\n\
If you specify them, the key is `INSTANCE.COMPONENT.ATTRIBUTE'\n\
and the class is `Emacs.CLASS.SUBCLASS'.")
  (attribute, class, component, subclass)
     Lisp_Object attribute, class, component, subclass;
{
  register char *value;
  char *name_key;
  char *class_key;

  CHECK_STRING (attribute, 0);
  CHECK_STRING (class, 0);

  if (!NILP (component))
    CHECK_STRING (component, 1);
  if (!NILP (subclass))
    CHECK_STRING (subclass, 2);
  if (NILP (component) != NILP (subclass))
    error ("x-get-resource: must specify both COMPONENT and SUBCLASS or neither");

  validate_x_resource_name ();

  /* Allocate space for the components, the dots which separate them,
     and the final '\0'.  Make them big enough for the worst case.  */
  name_key = (char *) alloca (XSTRING (Vx_resource_name)->size
			      + (STRINGP (component)
				 ? XSTRING (component)->size : 0)
			      + XSTRING (attribute)->size
			      + 3);

  class_key = (char *) alloca ((sizeof (EMACS_CLASS) - 1)
			       + XSTRING (class)->size
			       + (STRINGP (subclass)
				  ? XSTRING (subclass)->size : 0)
			       + 3);

  /* Start with emacs.FRAMENAME for the name (the specific one)
     and with `Emacs' for the class key (the general one).  */
  strcpy (name_key, XSTRING (Vx_resource_name)->data);
  strcpy (class_key, EMACS_CLASS);

  strcat (class_key, ".");
  strcat (class_key, XSTRING (class)->data);

  if (!NILP (component))
    {
      strcat (class_key, ".");
      strcat (class_key, XSTRING (subclass)->data);

      strcat (name_key, ".");
      strcat (name_key, XSTRING (component)->data);
    }

  strcat (name_key, ".");
  strcat (name_key, XSTRING (attribute)->data);

  value = x_get_string_resource (Qnil,
				 name_key, class_key);

  if (value != (char *) 0)
    return build_string (value);
  else
    return Qnil;
}

static void
InvertCursor (struct mac_output *mwp)
{
  Rect r;
  GLYPH glyph;
  int cs;
  int width = FONT_WIDTH (mwp->mFP->output_data.mac->font);
  int height = mwp->mLineHeight;

  /* Examine glyph at cursor position (default to ASCII space).  */
  if (mwp->mFP->current_glyphs->enable[curs_y] && curs_x < mwp->mFP->current_glyphs->used[curs_y])
    glyph = mwp->mFP->current_glyphs->glyphs[curs_y][curs_x];
  else
    glyph = SPACEGLYPH;

  r.left = kInsetPixels + curs_x * width;
  
  cs = CHAR_CHARSET (FAST_GLYPH_CHAR (glyph));
  /* Multiply width by number of columns if the cursor is on a
     multi-column character.  */
  width *= (cs == CHARSET_COMPOSITION
            ? cmpchar_table[COMPOSITE_CHAR_ID (glyph)]->width
            : CHARSET_WIDTH (cs));
  r.right = r.left + width;

  r.top = kInsetPixels + curs_y * height;
  r.bottom = r.top + height;
	
  SetPort (GetWindowPort(mwp->mWP));
	
  InvertRect (&r);
}

static void
CursorOn (struct mac_output *mwp)
{
  InvertCursor (mwp);
}

static void
CursorOff (struct mac_output *mwp)
{
  InvertCursor (mwp);  /* it's just a toggle */
}

void
mac_cursor_to (int row, int col)
{
  struct frame *f;
  int numCols, numRows;
  
  f = updating_frame;
  if (f == 0)
    f = selected_frame;

  numCols = f->output_data.mac->mNumCols;
  numRows = f->output_data.mac->mNumRows;

  if (termscript)
    fprintf (termscript, "<CURSOR TO %d, %d>\n", row, col);

  if (col < 0 || col >= numCols || row < 0 || row >= numRows)
    {
      SysBeep (10);
      ExitToShell ();
    }

  if (gUpdateInProgress)  /* temporarily change cursor position within an update */
    {
      curs_x = col;
      curs_y = row;
    }
  else  /* change cursor position in Frame structure */
    {
      if (!gFlagAppSuspend)  /* don't redraw cursor if Emacs is suspended */
        CursorOff (f->output_data.mac);

      curs_x = col;
      curs_y = row;
      FRAME_CURSOR_X (f) = col;
      FRAME_CURSOR_Y (f) = row;

      if (!gFlagAppSuspend)
        CursorOn (f->output_data.mac);
    }
}

void
mac_clear_to_end ()
{
  int numRows = updating_frame->output_data.mac->mNumRows;
  int numCols = updating_frame->output_data.mac->mNumCols;
  int i;

  if (termscript)
    fprintf (termscript, "<CLEAR TO END>\n");

  mac_clear_end_of_line (numCols);
  for (i = curs_y + 1; i < numRows; i++)
    {
      mac_cursor_to (i, 0);
      mac_clear_end_of_line (numCols);
    }
}

static Rect *GetWindowPortRect(WindowRef win)
{
  static Rect x; // do I have to worry about unexec here??
  GetWindowPortBounds(win,&x);
  return &x;
}

void
mac_clear_frame ()
{
  Rect windowRect;
  mac_output *mwp = updating_frame->output_data.mac;
	
  if (termscript)
    fprintf (termscript, "<CLEAR FRAME>\n");

  SetPort (GetWindowPort(mwp->mWP));

  GetWindowPortBounds(mwp->mWP,&windowRect);  //  windowRect = mwp->mWP->portRect;

  SetFrameFgBgColors (mwp->foreground_pixel, mwp->background_pixel);
  EraseRect (&windowRect);

  curs_x = 0;
  curs_y = 0;
}

/* Clear to the end of the line.
   Erase the current text line from the nominal cursor position (inclusive)
   to column FIRST_UNUSED (exclusive).  The idea is that everything
   from FIRST_UNUSED onward is already erased.  */
   
void
mac_clear_end_of_line (int first_unused)
{
  struct frame *f = updating_frame;
  mac_output *mwp = f->output_data.mac;
  int charWidth = mwp->mCharWidth;
  int lineHeight = mwp->mLineHeight;
  int numCols = mwp->mNumCols;
  Rect r;
	
  if (curs_y < 0 || curs_y >= f->height)
    return;
  if (first_unused <= 0)
    return;

  if (first_unused >= FRAME_WINDOW_WIDTH (f))
    first_unused = FRAME_WINDOW_WIDTH (f);

  first_unused += FRAME_LEFT_SCROLL_BAR_WIDTH (f);

  if (termscript)
    fprintf (termscript, "<CLEAR END OF LINE>\n");

  r.left = CHAR_TO_PIXEL_COL (f, curs_x);
  r.right = r.left + FONT_WIDTH (f->output_data.mac->font) * (first_unused - curs_x);
  r.top = CHAR_TO_PIXEL_ROW (f, curs_y);
  r.bottom = r.top + f->output_data.mac->mLineHeight;
	
  SetPort (GetWindowPort(mwp->mWP));

  SetFrameFgBgColors (mwp->foreground_pixel, mwp->background_pixel);
  EraseRect (&r);	
}

/* This one's definition is slightly tricky: scroll down n lines (if n
   is positive) or up -n lines (if n is negative).  The region of the
   window in which scrolling takes place is from the first line of the
   window to mTermWinSize, which is set by mac_set_window_size.  */
void
mac_ins_del_lines (int vpos, int n)
{
  mac_output *mwp = updating_frame->output_data.mac;
  int charWidth = mwp->mCharWidth;
  int lineHeight = mwp->mLineHeight;
  int numCols = mwp->mNumCols;
  int termWinSize = mwp->mTermWinSize;
  RgnHandle tempRegion;
  Rect scrollBox;
	
  if (termscript)
    fprintf (termscript, "<INS DEL LINES %d, %d>\n", vpos, n);

  /* sanity checks, probably unnecessary */
  if (n == 0)
    return;
		
  if (n > 0 && vpos + n > termWinSize)
    n = termWinSize - vpos;

  if (n < 0 && vpos - n > termWinSize)
    n = vpos - termWinSize;

  SetPort(GetWindowPort(mwp->mWP));
  tempRegion = NewRgn ();

  scrollBox.left = kInsetPixels;
  scrollBox.right = kInsetPixels + numCols * charWidth;
  scrollBox.top = kInsetPixels + vpos * lineHeight;
  scrollBox.bottom = kInsetPixels + termWinSize * lineHeight;
	
  SetFrameFgBgColors (mwp->foreground_pixel, mwp->background_pixel);

  /* Works when n is positive or negative! */
  ScrollRect (&scrollBox, 0, n * lineHeight, tempRegion);

  DisposeRgn (tempRegion);
}	

void
mac_reassert_line_highlight (int highlight, int vPos)
{
#pragma unused(vPos)

  struct frame *f;
  
  f = updating_frame;
  if (f == 0)
    f = selected_frame;

  f->output_data.mac->mHighlight = highlight;
  
  if (termscript)
    fprintf (termscript, "<REASSERT LINE HIGHLIGHT %d, %d>\n", highlight, vPos);
}

void
mac_change_line_highlight (int highlight, int vPos, int firstUnusedHPos)
{
  updating_frame->output_data.mac->mHighlight = highlight;
  
  mac_cursor_to (vPos, 0);
  mac_clear_end_of_line (firstUnusedHPos);

  if (termscript)
    fprintf (termscript, "<CHANGE LINE HIGHLIGHT %d, %d, %d>\n", highlight, vPos, firstUnusedHPos);
}

/* Structure borrowed from Xlib.h to represent two-byte characters in
   dumpglyphs.  */
typedef struct {
  unsigned char byte1;
  unsigned char byte2;
} XChar2b;

/* Glyph display code adapted from xterm.c and w32term.c.

   Display a sequence of N glyphs found at GP.
   F is the frame to output to.  LEFT and TOP are starting coords.
   HL is 1 if this text is highlighted, 2 if the cursor is on it,
   3 if should appear in its mouse-face.
   JUST_FOREGROUND if 1 means draw only the foreground;
   don't alter the background.

   CMPCHARP if non NULL is a pointer to the struct cmpchar_info, which
   means drawing glyphs on the same column.  This is set to non NULL
   only when recursively called within dumpglyphs to draw a composite
   character specified by CMPCHAR.

   Since the display generation code is responsible for calling
   compute_char_face and compute_glyph_face on everything it puts in
   the display structure, we can assume that the face code on each
   glyph is a valid index into FRAME_COMPUTED_FACES (f), and the one
   to which we can actually apply intern_face.  Call this function
   with input blocked.

   Return overall pixel width of the drawn glyphs.  */

static int
dumpglyphs (f, left, top, gp, n, hl, just_foreground, cmpcharp)
     struct frame *f;
     int left, top;
     register GLYPH *gp; /* Points to first GLYPH. */
     register int n;  /* Number of glyphs to display. */
     int hl;
     int just_foreground;
     struct cmpchar_info *cmpcharp;
{
  /* Holds characters to be displayed. */
  XChar2b *x_2byte_buffer
    = (XChar2b *) alloca (FRAME_WINDOW_WIDTH (f) * sizeof (*x_2byte_buffer));
  register XChar2b *cp;		/* Steps through x_2byte_buffer[]. */
  char *x_1byte_buffer
    = (char *) alloca (FRAME_WINDOW_WIDTH (f) * sizeof (*x_1byte_buffer));
  register int tlen = GLYPH_TABLE_LENGTH;
  register Lisp_Object *tbase = GLYPH_TABLE_BASE;
  WindowPtr window = f->output_data.mac->mWP;  /* QuickDraw window */
  int orig_left = left;
  int gidx = 0;
  int i;
  
  SetPort(GetWindowPort(window));

  while (n > 0)
    {
      /* Get the face-code of the next GLYPH.  */
      int cf, len;
      GLYPH g = *gp;
      int ch, charset;
      Lisp_Object first_ch;
      /* HIGHEST and LOWEST are used while drawing a composite
         character.  The meanings are described later.  */
      int highest, lowest;

      GLYPH_FOLLOW_ALIASES (tbase, tlen, g);
      cf = (cmpcharp ? cmpcharp->face_work : FAST_GLYPH_FACE (g));
      ch = FAST_GLYPH_CHAR (g);
      if (unibyte_display_via_language_environment
	  && SINGLE_BYTE_CHAR_P (ch)
	  && (ch >= 0240
	      || (ch >= 0200 && !NILP (Vnonascii_translation_table))))
	ch = unibyte_char_to_multibyte (ch);
      if (gidx == 0) XSETFASTINT (first_ch, ch);
      charset = CHAR_CHARSET (ch);
      if (charset == CHARSET_COMPOSITION)
	{
	  /* We must draw components of the composite character on the
             same column.  */
	  cmpcharp = cmpchar_table[COMPOSITE_CHAR_ID (ch)];

	  /* Set the face in the slot for work.  */
	  cmpcharp->face_work = cf;

	  /* We don't need the return value ... */
	  dumpglyphs (f, left, top, cmpcharp->glyph, cmpcharp->glyph_len,
		      hl, just_foreground, cmpcharp);
	  /* ... because the width of just drawn text can be
             calculated as follows.  */
	  left += FONT_WIDTH (f->output_data.mac->font) * cmpcharp->width;

	  ++gp, --n;
	  while (gp && (*gp & GLYPH_MASK_PADDING)) ++gp, --n;
	  cmpcharp = NULL;
	  continue;
	}

      /* Find the run of consecutive glyphs which can be drawn with
	 the same GC (i.e. the same charset and the same face-code).
	 Extract their character codes into X_2BYTE_BUFFER.  If
	 CMPCHARP is not NULL, face-code is not checked because we use
	 only the face specified in `cmpcharp->face_work'.  */
      cp = x_2byte_buffer;
      while (n > 0)
	{
	  int this_charset, c1, c2;

	  g = *gp;
	  GLYPH_FOLLOW_ALIASES (tbase, tlen, g);
	  ch = FAST_GLYPH_CHAR (g);
	  if (unibyte_display_via_language_environment
	      && SINGLE_BYTE_CHAR_P (ch)
	      && (ch >= 0240
		  || (ch >= 0200 && !NILP (Vnonascii_translation_table))))
	    ch = unibyte_char_to_multibyte (ch);
	  SPLIT_CHAR (ch, this_charset, c1, c2);
	  if (this_charset != charset
	      || (cmpcharp == NULL && FAST_GLYPH_FACE (g) != cf))
	    break;

	  if (c2 > 0)
	    cp->byte1 = c1, cp->byte2 = c2;
	  else
	    cp->byte1 = 0, cp->byte2 = c1;
	  ++cp;
	  ++gp, --n;
	  while (gp && (*gp & GLYPH_MASK_PADDING))
	    ++gp, --n;
	}

      /* LEN gets the length of the run.  */
      len = cp - x_2byte_buffer;
      /* Now output this run of chars, with the font and pixel values
	 determined by the face code CF.  */
      {
	struct face *face = FRAME_DEFAULT_FACE (f);
	MacFontStruct *font = NULL;
	/* Since the Mac does not have graphics contexts, we only keep
	   track of the foreground and background colors for
	   displaying text.  We will always call SetFrameFgBgColors
	   with these colors right before drawing any text.  */
	unsigned long foregroundColor, backgroundColor;
	int stippled = 0;
	int line_height = f->output_data.mac->mLineHeight;
	/* Pixel width of each glyph in this run.  */
	int glyph_width
	  = (FONT_WIDTH (f->output_data.mac->font)
	     * (cmpcharp ? cmpcharp->width : CHARSET_WIDTH (charset)));
	/* Overall pixel width of this run.  */
	int run_width
	  = (FONT_WIDTH (f->output_data.mac->font)
	     * (cmpcharp ? cmpcharp->width : len * CHARSET_WIDTH (charset)));
	/* A flag to tell if we have already filled background.  We
	   fill background in advance in the following cases:
	   1) A face has stipple.
	   2) A height of font is shorter than LINE_HEIGHT.
	   3) Drawing a composite character.
	   4) Font has non-zero _MULE_BASELINE_OFFSET property.
	   After filling background, we draw glyphs by XDrawString16.  */
	int background_filled;
	/* Baseline position of a character, offset from TOP.  */
	int baseline;
	/* The property value of `_MULE_RELATIVE_COMPOSE' and
           `_MULE_DEFAULT_ASCENT'.  */
	int relative_compose = 0, default_ascent = 0;
	/* 1 if we find no font or a font of inappropriate size.  */
	int require_clipping;

#if 0
	/* HL = 3 means use a mouse face previously chosen.  */
	if (hl == 3)
	  cf = FRAME_X_DISPLAY_INFO (f)->mouse_face_face_id;
#endif

	/* First look at the face of the text itself.  */
	if (cf != 0)
	  {
	    /* It's possible for the display table to specify
	       a face code that is out of range.  Use 0 in that case.  */
	    if (cf < 0 || cf >= FRAME_N_COMPUTED_FACES (f)
		|| FRAME_COMPUTED_FACES (f) [cf] == 0)
	      cf = 0;

	    if (cf == 1)
	      face = FRAME_MODE_LINE_FACE (f);
	    else
	      face = intern_face (f, FRAME_COMPUTED_FACES (f) [cf]);
	    if (FACE_STIPPLE (face))
	      stippled = 1;
	  }

	/* Then comes the distinction between modeline and normal text.  */
	else if (hl == 0)
	  ;
	else if (hl == 1)
	  {
	    face = FRAME_MODE_LINE_FACE (f);
	    if (FACE_STIPPLE (face))
	      stippled = 1;
	  }

	/* Setting appropriate font and gc for this charset.  */
	if (charset != CHARSET_ASCII)
	  {
	    int font_id;
	    int fontset = FACE_FONTSET (face);
	    struct font_info *fontp;

	    if ((fontset < 0 && (fontset = FRAME_FONTSET (f)) < 0)
		|| !(fontp = FS_LOAD_FONT (f, FRAME_MAC_FONT_TABLE (f),
					   charset, NULL, fontset)))
	      goto font_not_found;

	    font = (MacFontStruct *) (fontp->font);
	    /* choosing a font on the Mac amounts to making these three calls */
	    TextFont (font->mFontNum);
	    TextSize (font->mFontSize);
	    TextFace (font->mFontFace);
	    foregroundColor = face->foreground;
	    backgroundColor = face->background;
	    
	    baseline
	      = (font->mTwoByte
		 ? (line_height + font->mAscent - font->mDescent) / 2
		 : f->output_data.mac->font_baseline - fontp->baseline_offset);
	    if (FONT_HEIGHT (font) <= line_height
		&& (font->mAscent > baseline
		    || font->mDescent > line_height - baseline))
	      /* Adjust baseline for this font to show the whole
                 glyphs in a line.  */
	      baseline = line_height - font->mDescent;
	      
	    if (cmpcharp && cmpcharp->cmp_rule == NULL)
	      {
		relative_compose = fontp->relative_compose;
		default_ascent = fontp->default_ascent;
	      }

	    /* We have to change code points in the following cases.  */
	    if (fontp->font_encoder)
	      {
		/* This font requires CCL program to calculate code
                   point of characters.  */
		struct ccl_program *ccl = fontp->font_encoder;

		if (CHARSET_DIMENSION (charset) == 1)
		  for (cp = x_2byte_buffer; cp < x_2byte_buffer + len; cp++)
		    {
		      ccl->reg[0] = charset;
		      ccl->reg[1] = cp->byte2;
		      ccl_driver (ccl, NULL, NULL, 0, 0, NULL);
		      /* We assume that MSBs are appropriately
                         set/reset by CCL program.  */
		      if (!font->mTwoByte)	/* 1-byte font */
			cp->byte1 = 0, cp->byte2 = ccl->reg[1];
		      else
			cp->byte1 = ccl->reg[1], cp->byte2 = ccl->reg[2];
		    }
		else
		  for (cp = x_2byte_buffer; cp < x_2byte_buffer + len; cp++)
		    {
		      ccl->reg[0] = charset;
		      ccl->reg[1] = cp->byte1, ccl->reg[2] = cp->byte2;
		      ccl_driver (ccl, NULL, NULL, 0, 0, NULL);
		      /* We assume that MSBs are appropriately
                         set/reset by CCL program.  */
		      if (!font->mTwoByte)	/* 1-byte font */
			cp->byte1 = 0, cp->byte2 = ccl->reg[1];
		      else
			cp->byte1 = ccl->reg[1], cp->byte2 = ccl->reg[2];
		    }
	      }
	    else if (fontp->encoding[charset])
	      {
		int enc = fontp->encoding[charset];

		if ((enc == 1 || enc == 2) && CHARSET_DIMENSION (charset) == 2)
		  for (cp = x_2byte_buffer; cp < x_2byte_buffer + len; cp++)
		    cp->byte1 |= 0x80;
		if (enc == 1 || enc == 3)
		  for (cp = x_2byte_buffer; cp < x_2byte_buffer + len; cp++)
		    cp->byte2 |= 0x80;
                /* Japanese fonts on the Mac use the SJIS coding,
                   which is encoded by the macro ENCODE_SJIS as a
                   special case and not by a CCL program.  The
                   encoding field is set in mac_load_font to indicate
                   this (same technique used in w32term.c).  */
                if (enc == 4)
                  {
                    if (CHARSET_DIMENSION (charset) == 2)
                      {
                        int sjis1, sjis2;
                        for (cp = x_2byte_buffer; cp < x_2byte_buffer + len; cp++)
                          {
                            ENCODE_SJIS (cp->byte1, cp->byte2, sjis1, sjis2);
                            cp->byte1 = sjis1;
                            cp->byte2 = sjis2;
                          }
                      }
                    else
                      for (cp = x_2byte_buffer; cp < x_2byte_buffer + len; cp++)
                        cp->byte2 |= 0x80;
                  }
	      }
	  }
	else
	  {
	  font_not_found:
	    if (charset == CHARSET_ASCII || charset == charset_latin_iso8859_1)
	      {
		font = FACE_FONT (face);
		if (!font || font == (MacFontStruct *) FACE_DEFAULT)
		  font = f->output_data.mac->font;
		baseline = FONT_BASE (f->output_data.mac->font);
#if 0  /* forget this check on the Mac */
		if (charset == charset_latin_iso8859_1)
		  {
		    if (font->max_char_or_byte2 < 0x80)
		      /* This font can't display Latin1 characters.  */
		      font = NULL;
		    else
		      {
			for (cp = x_2byte_buffer; cp < x_2byte_buffer + len; cp++)
			  cp->byte2 |= 0x80;
		      }
		  }
#endif
	      }
	    
	    foregroundColor = face->foreground;
	    backgroundColor = face->background;
	    if (font)
	      {
	        TextFont (font->mFontNum);
	        TextSize (font->mFontSize);
	        TextFace (font->mFontFace);
	      }
	  }

	/* Now override that if the cursor's on this character.  */
	if (hl == 2)
	  {
	    /* The cursor overrides stippling.  */
	    stippled = 0;

#if 0	    
            if (font == f->output_data.mac->font
		&& face->background == f->output_data.mac->background_pixel
		&& face->foreground == f->output_data.mac->foreground_pixel
		&& !cmpcharp)
	      {
		gc = f->output_data.mac->cursor_gc;
	      }
	    /* Cursor on non-default face: must merge.  */
	    else
#endif
	      {
		backgroundColor = f->output_data.mac->cursor_pixel;
		foregroundColor = face->background;
		/* If the glyph would be invisible,
		   try a different foreground.  */
		if (foregroundColor == backgroundColor)
		  foregroundColor = face->foreground;
		/* Make sure the cursor is distinct from text in this face.  */
		if (backgroundColor == face->background
		    && foregroundColor == face->foreground)
		  {
		    backgroundColor = face->foreground;
		    foregroundColor = face->background;
		  }
		if (font)
		  {
		    TextFont (font->mFontNum);
	            TextSize (font->mFontSize);
	            TextFace (font->mFontFace);
                  }
	      }
	  }

	if (font)
	  require_clipping = (!NILP (Vclip_large_size_font)
			      && (font->mAscent > baseline
				  || font->mDescent > line_height - baseline
				  || (!cmpcharp
				      && FONT_WIDTH (font) > glyph_width)));

	if (font && (just_foreground || (cmpcharp && gidx > 0)))
	  background_filled = 1;
	else if (stippled)
	  {
	    background_filled = 1;
	  }
	else if (!font
		 || FONT_BASE (font) < baseline
		 || (FONT_HEIGHT (font) - FONT_BASE (font)
		     < line_height - baseline)
		 || FONT_WIDTH (font) < glyph_width
		 || cmpcharp)
	  {
            Rect r;

            SetFrameFgBgColors (foregroundColor, backgroundColor);

  	    r.left = left;
  	    r.top = top;
  	    r.right = left + run_width;
  	    r.bottom = top + line_height;
  	    
  	    EraseRect (&r);

	    background_filled = 1;
	    if (cmpcharp)
	      /* To assure not to fill background while drawing
		 remaining components.  */
	      just_foreground = 1;
	  }
	else
	  background_filled = 0;

	if (font)
	  {
	    if (!cmpcharp)
	      {
                SetFrameFgBgColors (foregroundColor, backgroundColor);

		if (background_filled)
		  TextMode (srcOr);
		else
		  TextMode (srcCopy);
			
		if (require_clipping || FONT_WIDTH (font) != glyph_width)
		  {
		    /* setup the rectangular clipping region and draw each character */
		    RgnHandle oldClipRegion = NewRgn ();
		    GetClip (oldClipRegion);
		    for (i = 0; i < len; i++)
		      {
                        Rect clipRect;
		        clipRect.left = left + glyph_width * i;
                        clipRect.right = clipRect.left + glyph_width;
                        clipRect.top = top;
                        clipRect.bottom = top + line_height;
		        ClipRect (&clipRect);
                        MoveTo (left + glyph_width * i, top + baseline);
                        DrawText (x_2byte_buffer + i, 0, 2);
		      }
                    SetClip (oldClipRegion);
                    DisposeRgn (oldClipRegion);
                  }
		else
		  {
		    MoveTo (left, top + baseline);
		    /* See if this whole buffer can be output as 8-bit chars.
		       If so, copy x_2byte_buffer to x_1byte_buffer
		       and do it as 8-bit chars.  */
		    for (i = 0; i < len; i++)
		      {
			if (x_2byte_buffer[i].byte1 != 0)
			  break;
			x_1byte_buffer[i] = x_2byte_buffer[i].byte2;
		      }

		    if (i == len)
		      {
			DrawText (x_1byte_buffer, 0, len);
		      }
		    else
		      {
			/* We can't output them as 8-bit chars,
			   so do it as 16-bit chars.  */
		        for (i = 0; i < len; i++)
			  if (x_2byte_buffer[i].byte1 == 0)
			    DrawChar (x_2byte_buffer[i].byte2);
			  else
			    DrawText (&x_2byte_buffer[i], 0, 2);
		      }
		  }
	      }
	    else
	      {
                /* Handle composite characters (from w32term.c).  */
                int char_width = 0;

                if ((cmpcharp->cmp_rule || relative_compose)
                    && gidx == 0)
		  {
		    /* This is the first character.  Initialize variables.
		       HIGHEST is the highest position of glyphs ever
		       written, LOWEST the lowest position.  */
		    int xoffset = 0;
                    int yoffset = baseline;
                    int start = 0;

		    if (default_ascent
			&& CHAR_TABLE_P (Vuse_default_ascent)
			&& !NILP (Faref (Vuse_default_ascent, first_ch)))
		      {
			highest = default_ascent;
			lowest = 0;
		      }
                    /* TODO: per char metrics for Truetype and BDF
                       fonts.  */
		      {
                        highest = FONT_BASE (font) + 1;
                        lowest = - (FONT_HEIGHT (font) - FONT_BASE (font));
		      }

		    if (cmpcharp->cmp_rule)
		      xoffset = (int)(cmpcharp->col_offset[0]
				  * FONT_WIDTH (FRAME_FONT (f)));

                    i = 1;

                    /* Don't let characters go beyond the glyph
                       boundary whatever their over/underhangs. */
                    if (xoffset > glyph_width - char_width)
                      xoffset = glyph_width - char_width;

                    if (xoffset < 0)
                      xoffset = 0;

		    /* Draw the first character at the normal
                       position.  */
                    {
                      SetFrameFgBgColors (foregroundColor, backgroundColor);
                        
		      if (background_filled)
		        TextMode (srcOr);
		      else
		        TextMode (srcCopy);
			
		      MoveTo (left + xoffset, top + yoffset);
		      
		      if (x_2byte_buffer[i].byte1 == 0)
			DrawChar (x_2byte_buffer[start].byte2);
		      else
			DrawText (&x_2byte_buffer[start], 0, 2);
	
                      gidx++;
                    }
                  }
                else
                  i = 0;

                for (; i < len; i++, gidx++)
		  {
		    int xoffset = 0, yoffset = FONT_BASE (font);

		    if (relative_compose)
		      {
			/* No per char metrics on w32.  */
			if (NILP (Vignore_relative_composition)
			    || NILP (Faref (Vignore_relative_composition,
					    make_number (cmpcharp->glyph[gidx]))))
			  {
			    if (- (FONT_HEIGHT (font) - FONT_BASE (font))
                                >= relative_compose)
			      {
				/* Draw above the current glyphs.  */
				yoffset = highest + FONT_HEIGHT (font);
				highest += FONT_HEIGHT (font);
			      }
			    else if (FONT_BASE (font) <= 0)
			      {
				/* Draw beneath the current glyphs.  */
				yoffset = lowest;
				lowest -= FONT_HEIGHT (font);
			      }
			  }
			else
			  {
			    /* Draw the glyph at normal position.  If
                               it sticks out of HIGHEST or LOWEST,
                               update them appropriately.  */
			    if (FONT_BASE (font) > highest)
			      highest = FONT_BASE (font);
			    else if (- (FONT_HEIGHT (font) - FONT_BASE (font))
                                     < lowest)
			      lowest = - (FONT_HEIGHT (font) -
                                          FONT_BASE (font));
			  }
		      }
		    else if (cmpcharp->cmp_rule)
		      {
			int gref = (cmpcharp->cmp_rule[gidx] - 0xA0) / 9;
			int nref = (cmpcharp->cmp_rule[gidx] - 0xA0) % 9;
			int bottom, top;

			/* Re-encode GREF and NREF so that they specify
			   only Y-axis information:
			   0:top, 1:base, 2:bottom, 3:center  */
			gref = gref / 3 + (gref == 4) * 2;
			nref = nref / 3 + (nref == 4) * 2;

                        /* No per char metrics on w32.  */
			bottom = ((gref == 0 ? highest : gref == 1 ? 0
				   : gref == 2 ? lowest
				   : (highest + lowest) / 2)
			  - (nref == 0 ? FONT_HEIGHT (font)
			   : nref == 1 ? (FONT_HEIGHT (font) -
                                          FONT_BASE (font))
                                     : nref == 2 ? 0
                           : (FONT_HEIGHT (font) / 2)));
			top = bottom + FONT_HEIGHT (font);

			if (top > highest)
			  highest = top;
			if (bottom < lowest)
			  lowest = bottom;
			yoffset = bottom + FONT_HEIGHT (font);
			xoffset = (int)(cmpcharp->col_offset[gidx]
				    * FONT_WIDTH (FRAME_FONT(f)));
		      }

                    /* Don't let characters go beyond the glyph
                       boundary whatever their over/underhangs. */
                    if (xoffset > glyph_width - char_width)
                      xoffset = glyph_width - char_width;

                    if (xoffset < 0)
                      xoffset = 0;

                    {
                      SetFrameFgBgColors (foregroundColor, backgroundColor);
                        
		      if (background_filled)
		        TextMode (srcOr);
		      else
		        TextMode (srcCopy);
			
		      MoveTo (left + xoffset, top + yoffset);
		      
		      if (x_2byte_buffer[i].byte1 == 0)
			DrawChar (x_2byte_buffer[i].byte2);
		      else
			DrawText (&x_2byte_buffer[i], 0, 2);
		    }
		  }
	      }
	  }
	if (!font)
	  {
	    /* Show hollow rectangles to indicate that we found no font.  */
	    int limit = cmpcharp ? 1 : len;

	    for (i = 0; i < limit; i++)
	      {
	        Rect r;
	        
	        r.left = left + glyph_width * i;
	        r.top = top;
	        r.right = r.left + glyph_width - 1;
	        r.bottom = r.top + line_height - 1;

                SetFrameFgBgColors (foregroundColor, backgroundColor);
                
                EraseRect (&r);
	        FrameRect (&r);
	      }
	  }
	else if (require_clipping && !NILP (Vhighlight_wrong_size_font))
	  {
	    /* Show small 'L' shape to lower left of character to
               indicate that we found a font of inappropriate size.  */
	    int limit = cmpcharp ? 1 : len;

            SetFrameFgBgColors (foregroundColor, backgroundColor);
            
	    for (i = 0; i < limit; i++)
	      {
		MoveTo (left + glyph_width * i, top + line_height - 1);
		LineTo (left + glyph_width * i + 1, top + line_height - 1);
		MoveTo (left + glyph_width * i, top + line_height - 3);
		LineTo (left + glyph_width * i, top + line_height - 1);
	      }
	  }

	/* We should probably check for XA_UNDERLINE_POSITION and
	   XA_UNDERLINE_THICKNESS properties on the font, but let's
	   just get the thing working, and come back to that.  */
	{
	  /* Setting underline position based on the metric of the
	     current font results in shaky underline if it strides
	     over different fonts.  So, we set the position based only
	     on the default font of this frame.  */
	  int underline_position = f->output_data.mac->font_baseline + 1;

	  if (underline_position >= line_height)
	    underline_position = line_height - 1;
	  
	  if (face->underline)
	    {
	      Rect r;
	        
	      r.left = left;
	      r.top = top + underline_position;
	      r.right = r.left + run_width;
	      r.bottom = r.top + 1;
	      
	      PaintRect (&r);
	    }
	}

	if (!cmpcharp)
	  left += run_width;
      }
    }

  return (left - orig_left);
}

/* Output some text at the nominal frame cursor position.
   Advance the cursor over the text.
   Output LEN glyphs at START.

   `highlight', set up by XTreassert_line_highlight or
   XTchange_line_highlight, controls the pixel values used for
   foreground and background.  */

static void
mac_write_glyphs (start, len)
     register GLYPH *start;
     int len;
{
  register int temp_length;
  int mask;
  struct frame *f;

  BLOCK_INPUT;

  if (termscript)
    {
      int i;
      
      fprintf (termscript, "<WRITE GLYPHS ");
      for (i = 0; i < len; i++)
        fputc (FAST_GLYPH_CHAR (start[i]), termscript);
      fprintf (termscript, ", %d>\n", len);
#if 0
      fprintf (termscript, "<------FACES- ");
      for (i = 0; i < len; i++)
        fprintf (termscript, "%d", FAST_GLYPH_FACE (start[i]));
      fprintf (termscript, ">\n");
#endif
    }

  f = updating_frame;
  if (f == 0)
    {
      f = selected_frame;
      /* If not within an update, output at the frame's visible
	 cursor.  */
      curs_x = FRAME_CURSOR_X (f);
      curs_y = FRAME_CURSOR_Y (f);
    }

  if (!gUpdateInProgress && !gFlagAppSuspend)  /* skip cursor redraw if updating or suspended */
    CursorOff(f->output_data.mac);

  dumpglyphs (f,
	      CHAR_TO_PIXEL_COL (f, curs_x),
	      CHAR_TO_PIXEL_ROW (f, curs_y),
	      start, len, f->output_data.mac->mHighlight, 0, NULL);

  /* If we drew on top of the cursor, note that it is turned off.  */
  if (curs_y == f->phys_cursor_y
      && curs_x <= f->phys_cursor_x
      && curs_x + len > f->phys_cursor_x)
    f->phys_cursor_on = 0;

  curs_x += len;
  if (curs_x >= FRAME_CURSOR_X_LIMIT (f))
    curs_x = FRAME_CURSOR_X_LIMIT (f) - 1;

  if (!gUpdateInProgress && !gFlagAppSuspend)
    CursorOn (f->output_data.mac);      

  UNBLOCK_INPUT;
}

void
mac_ring_bell ()
{
  if (termscript)
    fprintf (termscript, "<RING BELL>\n");

  SysBeep (1);
}

/* Need empty hooks; otherwise term.c prints termcap control codes to
   stdout. */
void
mac_set_terminal_modes ()
{
  if (termscript)
    fprintf(termscript, "<SET TERMINAL MODES>\n");
}

void
mac_reset_terminal_modes()
{
  if (termscript)
    fprintf (termscript, "<RESET TERMINAL MODES>\n");
}

void
mac_update_begin (struct frame *f)
{
  if (termscript)
    fprintf (termscript, "<UPDATE BEGIN ====>>>> >\n");

  f->output_data.mac->mHighlight = 0;
  f->output_data.mac->mTermWinSize = f->output_data.mac->mNumRows;

  gUpdateInProgress = 1;

  if (!gFlagAppSuspend)
    CursorOff (f->output_data.mac);

  curs_x = FRAME_CURSOR_X (f);
  curs_y = FRAME_CURSOR_Y (f);
}

void
mac_update_end (struct frame *f)
{
  if (termscript)
    fprintf (termscript, "<UPDATE END <<<<==== >\n");

  gUpdateInProgress = 0;

  curs_x = FRAME_CURSOR_X (f);
  curs_y = FRAME_CURSOR_Y (f);

  if (!gFlagAppSuspend)
    CursorOn (f->output_data.mac);
}

void
mac_set_terminal_window (int size)
{
  if (termscript)
    fprintf (termscript, "<SET TERMINAL WINDOW %d>\n", size);

  updating_frame->output_data.mac->mTermWinSize = size;
}

void
mac_frame_raise_lower (struct frame *f, int raise)
{
  BLOCK_INPUT;
  if (raise)
    SelectWindow (f->output_data.mac->mWP);
  else
    SendBehind (f->output_data.mac->mWP, nil);
  UNBLOCK_INPUT;
}

/* Qheight and Qwidth are defined in frame.c.  */
extern Lisp_Object Qheight, Qwidth;

/* Qfont is defined in textprop.c.  */
extern Lisp_Object Qfont;

/* Mostly copied from x_set_frame_parameters in xfns.c */
void
mac_set_frame_parameters (f, alist)
     FRAME_PTR f;
     Lisp_Object alist;
{
  Lisp_Object tail;
  int length = XINT (Flength (alist));
  int i;
  Lisp_Object *parms = (Lisp_Object *) alloca (length * sizeof (Lisp_Object));
  Lisp_Object *values = (Lisp_Object *) alloca (length * sizeof (Lisp_Object));
  int redraw;
  extern unsigned long load_color ();
  /* If both of these parameters are present, it's more efficient to
     set them both at once.  So we wait until we've looked at the
     entire list before we set them.  */
  int width, height;
  struct gcpro gcpro1, gcpro2;
  redraw = 1;
  /* Extract parm names and values into those vectors.  */
  i = 0;
  for (tail = alist; CONSP (tail); tail = Fcdr (tail))
    {
      Lisp_Object elt;

      elt = Fcar (tail);
      parms[i] = Fcar (elt);
      CHECK_SYMBOL (parms[i], 1);
      values[i] = Fcdr (elt);
      i++;
    }

  GCPRO2 (*parms, *values);
  gcpro1.nvars = i;
  gcpro2.nvars = i;

  /* Provide default values for HEIGHT and WIDTH.  */
  if (FRAME_NEW_WIDTH (f))
    width = FRAME_NEW_WIDTH (f);
  else
    width = FRAME_WIDTH (f);

  if (FRAME_NEW_HEIGHT (f))
    height = FRAME_NEW_HEIGHT (f);
  else
    height = FRAME_HEIGHT (f);

  /* Now process them in reverse of specified order.  */
  for (i--; i >= 0; i--)
    {
      Lisp_Object prop = parms[i];
      Lisp_Object val  = values[i];

      store_frame_param (f, prop, val);

      if (EQ (prop, Qwidth) && NUMBERP (val))
 	width = XFASTINT (val);
      else if (EQ (prop, Qheight) && NUMBERP (val))
	height = XFASTINT (val);
      else if (EQ (prop, Qforeground_color))
	{
	  unsigned long new_color = load_color (f, val);
	  if (new_color != ~0)
	    {
	      FRAME_FOREGROUND_PIXEL (f) = new_color;
              /* To correctly display characters, when the foreground color
                 is changed, the foreground (background) color of the default
                 (mode_line) faces must also be changed.  This bug fix was due
                 to Josh Carter.  */
              FRAME_DEFAULT_PARAM_FACE (f)->foreground = new_color;
              FRAME_MODE_LINE_PARAM_FACE (f)->background = new_color;
              f->output_data.mac->cursor_pixel = new_color;
	      redraw = 1;
	      if (termscript)
		fprintf (termscript, "<FGCOLOR %lu>\n", new_color);
	    }
	}
      else if (EQ (prop, Qbackground_color))
	{
	  unsigned long new_color = load_color (f, val);
	  if (new_color != ~0)
	    {
	      FRAME_BACKGROUND_PIXEL (f) = new_color;
              /* Also set default face params.  See previous comment.  */
              FRAME_DEFAULT_PARAM_FACE (f)->background = new_color;
              FRAME_MODE_LINE_PARAM_FACE (f)->foreground = new_color;
	      redraw = 1;
	      if (termscript)
		fprintf (termscript, "<BGCOLOR %lu>\n", new_color);
	    }
	}
      else if (EQ (prop, Qtitle) && !EQ(val,Qnil))
	{
	  Str255 windowTitle;
	  if (strlen (XSTRING (val)->data) < 255)
	    {
	      strcpy (windowTitle, XSTRING (val)->data);
	      c2pstr (windowTitle);
	      SetWTitle (f->output_data.mac->mWP, windowTitle);
	    }
	  if (termscript)
	    fprintf (termscript, "<TITLE: %s>\n", XSTRING (val)->data);
	}
      else if (EQ (prop, intern ("reverse")) && EQ (val, Qt))
	{
	  unsigned long fg = FRAME_FOREGROUND_PIXEL (f);

	  FRAME_FOREGROUND_PIXEL (f) = FRAME_BACKGROUND_PIXEL (f);
	  FRAME_BACKGROUND_PIXEL (f) = fg;
	  if (termscript)
	    fprintf (termscript, "<INVERSE-VIDEO>\n");
	}
      else if (EQ (prop, Qfont))
        mac_set_frame_fontset (f, val, NULL);
    }

  {
    Lisp_Object frame;

    check_frame_size (f, &height, &width);

    XSETFRAME (frame, f);

    if (width != FRAME_WIDTH (f)
	|| height != FRAME_HEIGHT (f)
	|| FRAME_NEW_HEIGHT (f) || FRAME_NEW_WIDTH (f))
      Fset_frame_size (frame, make_number (width), make_number (height));
  }

  if (redraw) // && FRAME_MAC_P(f) && f->output_data.mac)
    {
      extern void recompute_basic_faces (FRAME_PTR);
      extern void redraw_frame (FRAME_PTR);

      recompute_basic_faces (f);
      if (f == selected_frame)
          redraw_frame (f);
    }

  UNGCPRO;
}

/* Routines made necessary by defining WINDOW_FRAME_P as WINDOW_MAC_P
   in frame.h.  Mostly copied from various emacs files.  */

/* CHANGE_GRAVITY is 1 when calling from Fset_frame_position, to
   really change the position, and 0 when calling from
   x_make_frame_visible (in that case, XOFF and YOFF are the current
   position values).  It is -1 when calling from
   x_set_frame_parameters, which means, do adjust for borders but
   don't change the gravity.  */

void
x_set_offset (f, xoff, yoff, change_gravity)
     struct frame *f;
     register int xoff, yoff;
     int change_gravity;
{
  if (change_gravity > 0)
    {
      f->output_data.mac->top_pos = yoff;
      f->output_data.mac->left_pos = xoff;
    }

  BLOCK_INPUT;
  MoveWindow (f->output_data.mac->mWP, xoff, yoff, false);
  UNBLOCK_INPUT;
}

/* Call this to change the size of frame F's x-window.  If
   CHANGE_GRAVITY is 1, we change to top-left-corner window gravity
   for this size change and subsequent size changes.  Otherwise we
   leave the window gravity unchanged.  */

void
x_set_window_size (f, change_gravity, cols, rows)
     struct frame *f;
     int change_gravity;
     int cols, rows;
{
  mac_output *mwp = f->output_data.mac;
  
  SetPort(GetWindowPort(mwp->mWP));

  mwp->mNumCols = cols;	
  mwp->mNumRows = rows;

  /* pick up the new font width and height from the newly set font field */
  mwp->mCharWidth = mwp->font->mWidth;
  mwp->mLineHeight = mwp->font->mHeight;
  
  mwp->mHomeX = kInsetPixels;
  /* share the leading equally between above the line and below the line */
  mwp->mHomeY = kInsetPixels + mwp->font->mAscent + mwp->font->mLeading / 2;

  SizeWindow (mwp->mWP, mwp->mCharWidth * mwp->mNumCols + 2 * kInsetPixels, mwp->mLineHeight * mwp->mNumRows + 2 * kInsetPixels, false);
  ShowWindow (mwp->mWP);  

  change_frame_size (f, rows, cols, 0, 0);

  SET_FRAME_GARBAGED (f);
}

int
x_pixel_width (struct frame *f)
{
  return f->output_data.mac->mNumCols * f->output_data.mac->mCharWidth;
}

int
x_pixel_height (struct frame *f)
{
  return f->output_data.mac->mNumRows * f->output_data.mac->mLineHeight;
}

int
x_char_width (struct frame *f)
{
  return f->output_data.mac->mCharWidth;
}

int
x_char_height (struct frame *f)
{
  return f->output_data.mac->mLineHeight;
}

/* Insert a description of internally-recorded parameters of frame X
   into the parameter alist *ALISTPTR that is to be given to the user.
   Only parameters that are specific to W32 and whose values are not
   correctly recorded in the frame's param_alist need to be considered
   here.  */

void
x_report_frame_params (f, alistptr)
     struct frame *f;
     Lisp_Object *alistptr;
{
  Lisp_Object tem;

  /* Represent negative positions (off the top or left screen edge) in
     a way that Fmodify_frame_parameters will understand correctly.  */
  XSETINT (tem, f->output_data.mac->left_pos);
  if (f->output_data.mac->left_pos >= 0)
    store_in_alist (alistptr, Qleft, tem);
  else
    store_in_alist (alistptr, Qleft, Fcons (Qplus, Fcons (tem, Qnil)));

  XSETINT (tem, f->output_data.mac->top_pos);
  if (f->output_data.mac->top_pos >= 0)
    store_in_alist (alistptr, Qtop, tem);
  else
    store_in_alist (alistptr, Qtop, Fcons (Qplus, Fcons (tem, Qnil)));
}

void
x_iconify_frame (struct frame *f)
{
  /* not implemented for now */
}

void
x_make_frame_visible (struct frame *f)
{
  /* not implemented for now */
}

void
x_make_frame_invisible (struct frame *f)
{
  /* not implemented for now */
}

void
x_set_mouse_position (f, x, y)
     struct frame *f;
     int x, y;
{
  /* not implemented for now */
}

void
x_set_mouse_pixel_position (f, pix_x, pix_y)
     struct frame *f;
     int pix_x, pix_y;
{
  /* not implemented for now */
}

void
x_set_frame_parameters (f, alist)
     FRAME_PTR f;
     Lisp_Object alist;
{
  mac_set_frame_parameters (f, alist);
}

/* Make these functions instead of macros because it's hard to make all
   source files which use them include them. */
int
PIXEL_HEIGHT (struct frame *f)
{
  return x_pixel_height (f);
}

int
FRAME_INTERNAL_BORDER_WIDTH (struct frame *f)
{
  return kInsetPixels;
}

int
FRAME_LINE_HEIGHT (struct frame *f)
{
  return f->output_data.mac->mLineHeight;
}

void
x_destroy_window (f)
     struct frame *f;
{
  DisposeMacWindow (f->output_data.mac);
}

/* Will no-op work?  */
void
x_sync (f)
     void *f;
{
}

/* FRAME is used only to get a handle on the X display.  We don't pass
   the display info directly because we're called from frame.c, which
   doesn't know about that structure.  */
Lisp_Object
x_get_focus_frame (frame)
     struct frame *frame;
{
  WindowPtr wp;
  mac_output *mwp;
  Lisp_Object xfocus;

  wp = FrontWindow ();
  if (!wp)
    return Qnil;

  mwp = (mac_output *) GetWRefCon (wp);
				
  XSETFRAME (xfocus, mwp->mFP);
  return xfocus;
}

void
mac_mouse_position (struct frame **f, int i, Lisp_Object *obj,
		    enum scroll_bar_part *sbp,
		    Lisp_Object *x,
		    Lisp_Object *y,
		    unsigned long *time)
{
}

extern Lisp_Object Vwindow_system;

/* Given a pixel position (PIX_X, PIX_Y) on the frame F, return glyph
   co-ordinates in (*X, *Y).  Set *BOUNDS to the rectangle that the
   glyph at X, Y occupies, if BOUNDS != 0.  If NOCLIP is nonzero, do
   not force the value into range.  */
void
pixel_to_glyph_coords (f, pix_x, pix_y, x, y, bounds, noclip)
     FRAME_PTR f;
     register int pix_x, pix_y;
     register int *x, *y;
     int *bounds;
     int noclip;
{
  /* Support tty mode: if Vwindow_system is nil, behave correctly. */
  if (NILP (Vwindow_system))
    {
      *x = pix_x;
      *y = pix_y;
      return;
    }

  /* Arrange for the division in PIXEL_TO_CHAR_COL etc. to round down
     even for negative values.  */
  if (pix_x < 0)
    pix_x -= FONT_WIDTH (f->output_data.mac->font) - 1;
  if (pix_y < 0)
    pix_y -= f->output_data.mac->mLineHeight - 1;

  pix_x = PIXEL_TO_CHAR_COL (f, pix_x);
  pix_y = PIXEL_TO_CHAR_ROW (f, pix_y);

  if (bounds)
    {
      abort ();  /* should never be called with non-nil BOUNDS */
#if 0
      bounds->left = pix_x * f->output_data.mac->mCharWidth;
      bounds->top = pix_y * f->output_data.mac->mLineHeight;
      bounds->right  = bounds->left + f->output_data.mac->mCharWidth - 1;
      bounds->bottom = bounds->top + f->output_data.mac->mLineHeight - 1;
#endif
    }

  if (!noclip)
    {
      if (pix_x < 0)
	pix_x = 0;
      else if (pix_x > f->width)
	pix_x = f->width;

      if (pix_y < 0)
	pix_y = 0;
      else if (pix_y > f->height)
	pix_y = f->height;
    }

  *x = pix_x;
  *y = pix_y;
  
  /* This fixes the problem of Emacs incorrectly selecting the
     character to its immediate right when the mouse is clicked
     anywhere other than the first column of a multi-column character.  */
  if (*y < f->height && *x < f->width)
    while (*x >= 0 && f->current_glyphs->glyphs[*y][*x] & GLYPH_MASK_PADDING)
      (*x)--;
}

void
glyph_to_pixel_coords (f, x, y, pix_x, pix_y)
     FRAME_PTR f;
     register int x, y;
     register int *pix_x, *pix_y;
{
  *pix_x = CHAR_TO_PIXEL_COL (f, x);
  *pix_y = CHAR_TO_PIXEL_ROW (f, y);
}

/* set-window-configuration on window.c needs this.  (Copied from
   msdos.c.) */
void
x_set_menu_bar_lines (f, value, oldval)
     struct frame *f;
     Lisp_Object value, oldval;
{
  set_menu_bar_lines (f, value, oldval);
}

/* modify_event_symbol in keyboard.c needs this. */
char *
x_get_keysym_name (keysym)
     int keysym;
{
  return 0;
}

/* x_consider_frame_title in xdisp.c needs this */
void
x_implicitly_set_name (f, arg, oldval)
     FRAME_PTR f;
     Lisp_Object arg, oldval;
{
  /* no-op for now */
}

/* Table for translating Mac keycode to X keysym values.  Contributed
   by Sudhir Shenoy.  */
static unsigned char keyCodeToXKeySymTable[] = {
/* 0x00 - 0x3f */
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0,	0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 0x40 */
  0,			'\xae' /* kp. */,	0,			'\xaa' /* kp* */, 
  0,			'\xab' /* kp+ */,	0,			'\x7f' /* kp_clr */,
  0,			0,			0,			'\xaf' /* kp/ */,
  '\x8d' /* kp_ent */,	0,			'\xad' /* kp- */,	0,
/* 0x50 */
  0,			'\xbd' /* kp= */,	'\xb0' /* kp0 */,	'\xb1' /* kp1 */,
  '\xb2' /* kp2 */,	'\xb3' /* kp3 */,	'\xb4' /* kp4 */,	'\xb5' /* kp5 */,
  '\xb6' /* kp6 */,	'\xb7' /* kp7 */,	0,			'\xb8' /* kp8 */,
  '\xb9' /* kp9 */,	0,			0,			0,
/* 0x60 */
  '\xc2' /* F5 */,	'\xc3' /* F6 */,	'\xc4' /* F7 */,	'\xc0' /* F3 */,
  '\xc5' /* F8 */,	'\xc6' /* F9 */,	0,			'\xc8' /* F11 */, 
  0,			'\xca' /* F13 */,	0,			'\xcb' /* F14 */, 
  0,			'\xc7' /* F10 */,	0,			'\xc9' /* F12 */, 	
/* 0x70 */
  0,	 		'\xcc' /* F15 */,	'\x9e' /* ins */,	'\x95' /* home */,
  '\x9a' /* pgup */,	'\x9f' /* del */,	'\xc1' /* F4 */,	'\x9c' /* end */,
  '\xbf' /* F2 */,	'\x9b' /* pgdown */,	'\xbe' /* F1 */,	'\x51' /* left */,
  '\x53' /* right */,	'\x54' /* down */,	'\x52' /* up */,	0
};

int
KeyCodeToXKeySym (int keyCode, int *xKeySym)
{
  *xKeySym = keyCodeToXKeySymTable [keyCode & 0x7f];
  return *xKeySym != 0;
}

/* Emacs calls this whenever it wants to read an input event from the
   user. */
int
mac_read_socket (int sd, struct input_event *bufp, int numchars, int expected)
{
  EventRecord er;
  int theModifiers;

  if (numchars <= 0)
    abort ();

  /* Don't poll for events to process (specifically updateEvt) if
     window update currently already in progress.  A call to redisplay
     (in doWindowUpdate) can be preempted by another call to
     redisplay, causing blank regions to be left on the screen and the
     cursor to be left at strange places.  */
  if (gHandlingWindowUpdate)
    return 0;

  if (gDone)
    Fkill_emacs (make_number (1));

  if (WaitNextEvent (everyEvent, &er, (expected ? gAppSleepTime : 0L), NULL))
    switch (er.what)
      {
      case mouseDown:
      case mouseUp:
	{
	  WindowPtr	windowPtr;
	  SInt16	partCode;
	  SInt32	menuChoice;

	  partCode = FindWindow (er.where, &windowPtr);
					
	  switch (partCode)
	    {
	    case inMenuBar:
	      doAdjustMenus ();
	      menuChoice = MenuSelect (er.where);
	      doMenuChoice (menuChoice);
	      return 0;

	    case inContent:
	      if (windowPtr != FrontWindow ())
		SelectWindow (windowPtr);
	      bufp->kind = mouse_click;
	      bufp->code = 0;  /* only one mouse button */
							
	      switch (er.what)
		{
		case mouseDown:
		  bufp->modifiers = down_modifier;
		  break;
		case mouseUp:
		  bufp->modifiers = up_modifier;
		  break;
		}
							
	      {
		mac_output *mwp = (mac_output *) GetWRefCon (FrontWindow());
		Point mouseLoc = er.where;
								
		GlobalToLocal (&mouseLoc);
								
		bufp->x = mouseLoc.h;
		bufp->y = mouseLoc.v;

		XSETFRAME (bufp->frame_or_window, mwp->mFP);
								
		bufp->timestamp = er.when;
	      }
	      						
	      return 1;

	    case inDrag:
#if defined(TARGET_API_MAC_CARBON)
	      {
		Rect big = { SHRT_MIN,SHRT_MIN,SHRT_MAX,SHRT_MAX };
		DragWindow (windowPtr, er.where, &big);
	      }
#else
	      DragWindow (windowPtr, er.where, &qd.screenBits.bounds);
#endif
	      return 0;

	    case inGoAway:
	      if (TrackGoAway (windowPtr, er.where))
	        ;  /* Must be delete-frame.  (TI) */
		/* gDone = true; */
	      return 0;

	    /* window resize handling added --ben */
	    case inGrow:
	      doGrowWindow(windowPtr, &er);
	      return 0;
	    
	    /* window zoom handling added --ben */
	    case inZoomIn:
	    case inZoomOut:
	      if (TrackBox (windowPtr, er.where, partCode))
	        {
		  doZoomWindow (windowPtr, partCode);
		}
	      return 0;

	    default:
	      return 0;
	    }
	}
	break;
      case updateEvt:
      case osEvt:
      case activateEvt:
    	doEvents (&er);
	return 0;
      case keyDown:
      case autoKey:
	{
	  int keyCode = (er.message & keyCodeMask) >> 8;
	  int xKeySym;
	  
	  ObscureCursor ();

	  if (keyCode == 0x33)  /* delete key (charCode translated to 0x8) */
	    {
	      bufp->code = 0x7f;
	      bufp->kind = ascii_keystroke;
	    }
	  else if (KeyCodeToXKeySym (keyCode, &xKeySym))
	    {
	      bufp->code = 0xff00 | xKeySym;
	      bufp->kind = non_ascii_keystroke;
	    }	      
	  else
	    {
	      if (er.modifiers & (NILP (Vmac_command_key_is_meta) ? optionKey : cmdKey))
	        {
	          /* This code comes from Keyboard Resource, Appendix C of IM - Text.  This is
		     necessary since shift is ignored in KCHR table translation when option or
		     command is pressed. */
	          int newModifiers = er.modifiers & 0xf600;  /* mask off option and command */
	          int newKeyCode = keyCode | newModifiers;
	          Ptr KCHRPtr = (Ptr) GetScriptManagerVariable (smKCHRCache);
	          unsigned long someState = 0;
	          bufp->code = KeyTranslate (KCHRPtr, newKeyCode, &someState) & 0xff;
	        }
	      else
	        bufp->code = er.message & charCodeMask;
	      bufp->kind = ascii_keystroke;
	    }
	}

      /* If variable mac-convert-keyboard-input-to-latin-1 is non-nil,
         convert non-ASCII characters typed at the Mac keyboard
         (presumed to be in the Mac Roman encoding) to iso-latin-1
         encoding before they are passed to Emacs.  This enables the
         Mac keyboard to be used to enter non-ASCII iso-latin-1
         characters directly.  */
      if (mac_keyboard_text_encoding != kTextEncodingMacRoman && bufp->kind == ascii_keystroke && bufp->code >= 128)
	{
          static TECObjectRef converter = NULL;
          OSStatus theErr = noErr;
          OSStatus convertStatus = noErr;

          if (converter ==  NULL)
            {
              theErr = TECCreateConverter (&converter, kTextEncodingMacRoman, mac_keyboard_text_encoding);
              current_mac_keyboard_text_encoding = mac_keyboard_text_encoding;
            }
          else if (mac_keyboard_text_encoding != current_mac_keyboard_text_encoding)
            {
              /* Free the converter for the current encoding before
                 creating a new one.  */
              TECDisposeConverter (converter);
              theErr = TECCreateConverter (&converter, kTextEncodingMacRoman, mac_keyboard_text_encoding);
              current_mac_keyboard_text_encoding = mac_keyboard_text_encoding;
            } 
              
          if (theErr == noErr)
            {
              unsigned char ch = bufp->code;
              ByteCount actualInputLength, actualOutputLength;
              unsigned char outch;
                  
              convertStatus = TECConvertText (converter, &ch, 1, &actualInputLength,
                                                  &outch, 1, &actualOutputLength);
              if (convertStatus == noErr && actualInputLength == 1 && actualOutputLength == 1)
                bufp->code = outch;
            }
	}

	theModifiers = 0;
	if (er.modifiers & shiftKey)
	  theModifiers |= shift_modifier;
	if (er.modifiers & controlKey)
	  theModifiers |= ctrl_modifier;
	/* use option or command key as meta depending on value of mac-command-key-is-meta */
	if (er.modifiers & (NILP (Vmac_command_key_is_meta) ? optionKey : cmdKey))
	  theModifiers |= meta_modifier;
	bufp->modifiers	= theModifiers;
				
	{
	  mac_output *mwp = (mac_output *) GetWRefCon (FrontWindow ());
	  XSETFRAME (bufp->frame_or_window, mwp->mFP);
	}

	bufp->timestamp = er.when;

	return 1;

      case kHighLevelEvent:
        gDragAndDropFileList = Qnil;
        
        AEProcessAppleEvent(&er);
        
        /* Build a drag_n_drop type event as is done in
           constuct_drag_n_drop in w32term.c.  */
        if (!NILP (gDragAndDropFileList))
          {
            struct frame *f;
            WindowPtr wp;
            Lisp_Object frame;

            wp = FrontWindow ();
            if (!wp)
              f = NULL;
            else
              f = ((mac_output *) GetWRefCon (wp))->mFP;            
            
            bufp->kind = drag_n_drop;
            bufp->code = 0;
            bufp->timestamp = er.when;
            bufp->modifiers = 0;

            XSETINT (bufp->x, 0);
            XSETINT (bufp->y, 0);

            XSETFRAME (frame, f);
            bufp->frame_or_window = Fcons (frame, gDragAndDropFileList);

            /* Regardless of whether Emacs was suspended or in the
               foreground, ask it to redraw its entire screen.
               Otherwise parts of the screen can be left in an
               inconsistent state.  */
            if (wp)
	      InvalWindowRect(wp,GetWindowPortRect(wp));
//              InvalRect (&(wp->portRect));
            return 1;
          }
        else
          return 0;
        
      default:
	return 0;
      }
  else
    {
#if !defined(TARGET_API_MAC_CARBON)
      CheckAlarm ();  /* simulate the handling of a SIGALRM */
#endif
      return 0;
    }	
}

/* Handle drags in size box.  Based on code contributed by Ben
   Mesander and IM - Window Manager A.  */
static void
doGrowWindow (WindowPtr thisWindow, EventRecord *event)
{
  long growSize;
  Rect limitRect;
  int x, y;
  mac_output *mwp;

  SetRect(&limitRect, kMinDocSize, kMinDocSize, kMaxDocSize, kMaxDocSize);
  
  growSize = GrowWindow (thisWindow, event->where, &limitRect);

  /* see if it really changed size */
  if (growSize != 0)
    {
      mwp = (mac_output *) GetWRefCon (thisWindow);
      pixel_to_glyph_coords (mwp->mFP, LoWord (growSize), HiWord (growSize), &x, &y, 0, 1);
      x_set_window_size (mwp->mFP, 0, x, y);
      InvalWindowRect (thisWindow,GetWindowPortRect(thisWindow));//      InvalRect (&(thisWindow->portRect));
    }
}

/* Handle clicks in zoom box.  Calculation of "standard state" based
   on code in IM - Window Manager A and code contributed by Ben
   Mesander.  The standard state of an Emacs window is 80-characters
   wide (kNumCols) and as tall as will fit on the screen.  */
static void
doZoomWindow (WindowPtr thisWindow, int zoomInOrOut)
{
  GrafPtr savePort;
  Rect zoomRect, portRect;
  Point topLeft;
  int wTitleHeight, columns, rows, dummy, x, y;
  mac_output *mwp;
    
  mwp = (mac_output *) GetWRefCon (thisWindow);
   
  GetPort(&savePort);
  SetPort(GetWindowPort(thisWindow));
  EraseRect(GetWindowPortRect(thisWindow));
  //  EraseRect (&(thisWindow->portRect));  /* erase to avoid flicker */
  if (zoomInOrOut == inZoomOut)
    {
      SetPt(&topLeft, GetWindowPortRect(thisWindow)->left, GetWindowPortRect(thisWindow)->top);
      LocalToGlobal (&topLeft);

      /* calculate height of window's title bar */
#if defined(TARGET_API_MAC_CARBON)
      {
          Rect r;
          BitMap s;
          GetWindowBounds(thisWindow,kWindowStructureRgn,&r);
          wTitleHeight = topLeft.v - 1 - r.top + GetMBarHeight();
          
          /* get maximum height of window into zoomRect.bottom - zoomRect.top */
          GetQDGlobalsScreenBits(&s);
          zoomRect = s.bounds;
      }
#else
      wTitleHeight = topLeft.v - 1 - (**((WindowPeek) thisWindow)->strucRgn).rgnBBox.top + GetMBarHeight();

      /* get maximum height of window into zoomRect.bottom - zoomRect.top */
      zoomRect = qd.screenBits.bounds;
#endif
      zoomRect.top += wTitleHeight;
      InsetRect (&zoomRect, 3, 3);

      /* round its dimensions off to closest glyph units */
      pixel_to_glyph_coords (mwp->mFP, zoomRect.right - zoomRect.left, zoomRect.bottom - zoomRect.top, &dummy, &rows, 0, 1);

      /* make window as tall as possible and set to default width (kNumCols) */
      glyph_to_pixel_coords (mwp->mFP, kNumCols, rows, &x, &y);
      zoomRect.right = zoomRect.left + x;
      zoomRect.bottom = zoomRect.top + y;
#if defined(TARGET_API_MAC_CARBON)
      SetWindowStandardState(thisWindow,&zoomRect);
#else
      (**((WStateDataHandle) ((WindowPeek) thisWindow)->dataHandle)).stdState = zoomRect;
#endif
    }

  ZoomWindow (thisWindow, zoomInOrOut, thisWindow == FrontWindow());

  /* retrieve window size and update application values */
  GetWindowPortBounds(thisWindow,&portRect); //portRect = thisWindow->portRect;
  pixel_to_glyph_coords (mwp->mFP, portRect.right - portRect.left, portRect.bottom - portRect.top, &columns, &rows, 0, 1);
  x_set_window_size (mwp->mFP, 0, columns, rows);
  InvalWindowRect (thisWindow,&portRect);  /* cause a redraw */

  SetPort(savePort);
}

/* Need to override CodeWarrior's input function so no conversion is
   done on newlines Otherwise compiled functions in .elc files will be
   read incorrectly.  Defined in ...:MSL C:MSL
   Common:Source:buffer_io.c.  */
#ifdef __MWERKS__
void
__convert_to_newlines (unsigned char * p, size_t * n)
{
#pragma unused(p,n)
}

void
__convert_from_newlines (unsigned char * p, size_t * n)
{
#pragma unused(p,n)
}
#endif

/* Used by both MacKeyPending and MacGetChar.  MacKeyPending stores a
   key down or auto key event here for MacGetChar to return. */
static EventRecord theEvent;
 
/* Returns true if any keyboard key is pressed without retrieving the
   key used for stopping a loop by pressing any key.  */
int
MacKeyPending (void)
{
  if (WaitNextEvent (keyDownMask | autoKeyMask, &theEvent, 0L, NULL))
    switch (theEvent.what)
      {
      case keyDown:
      case autoKey:
	return 1;
      default:
	return 0;
      }
  else
    return 0;
}

/* Returns the keyboard character pressed when an ascii key is pressed
   used for console style menu selections for immediate actions.
   Fashioned after getch in SIOUX.c.  */
int MacGetChar (void)
{
  int c;
	
  if ((theEvent.message & keyCodeMask) == 0x00003300)
    return 127;		/* delete key */
	
  c = theEvent.message & charCodeMask;
	
  /* For meta keys to work, the global meta_keys must be set to 1 by
     adding the flag "km" to our termcap entry.  There are other ways
     but not as easy.  Also the default KCHR is such that once meta is
     pressed, the shift key doesn't have any effects: e.g., M-: is
     read as M-; .  */
	
  if ((theEvent.modifiers & cmdKey) && (theEvent.modifiers & shiftKey))
    switch (c)
      {
      case ';':
	c = ':';
	break;
      case ',':
	c = '<';
	break;
      case '.':
	c = '>';
	break;
      }

  /* Must return here since underscore has same char code as down arrow */
  if ((theEvent.modifiers & controlKey) && c == 0x1F)
    return c;

  switch (c)
    {
    case kLeftArrowCharCode:
      c = 'b' - 96;
      break;
    case kRightArrowCharCode:
      c = 'f' - 96;
      break;
    case kUpArrowCharCode:
      c = 'p' - 96;
      break;
    case kDownArrowCharCode:
      c = 'n' - 96;
      break;
    }

  if (theEvent.modifiers & cmdKey)
    c += 128;		/* meta key pressed */

  /* if (c == '.' && (theEvent.modifiers & cmdKey))
     exit(1); */
  
  return c;
}

/* set interprogram-paste-function to mac-paste-function in mac-win.el
   to enable Emacs to obtain the contents of the Mac clipboard. */
DEFUN ("mac-paste-function", Fmac_paste_function, Smac_paste_function, 0, 0, 0,
    "Return the contents of the Mac clipboard as a string.")
  ()
{
#if TARGET_API_MAC_CARBON
  ScrapRef scrap;
  UInt32 flavors,length;
  ScrapFlavorInfo *flavorInfo;
  int i;
  char *data;

  OSStatus status = GetCurrentScrap(&scrap);
  if (status != noErr)
      return Qnil;
#if 0
  status = GetScrapFlavorCount(scrap,&flavors);
  if (status != noErr)
      return Qnil;
  
  flavorInfo = alloca(sizeof(ScrapFlavorInfo)*flavors);
  if (!flavorInfo)
      return Qnil;

  status = GetScrapFlavorInfo(scrap,&flavors,flavorInfo);
  if (status != noErr)
      return Qnil;

  for (i=0;i<flavors;i++)
      if (flavorInfo[i].flavorType == kScrapFlavorTypeText)
          goto gotTextFlavor;
  return Qnil; // No text in scrap.

  gotTextFlavor:
#endif
  status = GetScrapFlavorSize(scrap,kScrapFlavorTypeText,&length);
  if (status != noErr || length == 0)
      return Qnil;

  data = alloca(length);
  if (!data)
      return Qnil;

  status = GetScrapFlavorData(scrap,kScrapFlavorTypeText,&length,data);
  if (status != noErr || length == 0)
      return Qnil;
  
  /* Emacs expects clipboard contents have Unix-style eol's */
  for (i = 0; i < length; i++)
    if (data[i] == '\r')
      data[i] = '\n';

  return make_string(data,length);
#else
  Lisp_Object value;
  Handle myHandle;
  long scrapOffset, rc, i;

  myHandle = NewHandle (0);  /* allocate 0-length data area */

  rc = GetScrap (myHandle, 'TEXT', &scrapOffset);
  if ( rc < 0 )
    return Qnil;

  HLock (myHandle);

  /* Emacs expects clipboard contents have Unix-style eol's */
  for (i = 0; i < rc; i++)
    if ((*myHandle)[i] == '\r')
      (*myHandle)[i] = '\n';

  value = make_string (*myHandle, rc);

  HUnlock (myHandle);
  
  DisposeHandle (myHandle);

  return value;
#endif
}

/* set interprogram-cut-function to mac-cut-function in mac-win.el
   to enable Emacs to write the top of the kill-ring to the Mac clipboard. */
DEFUN ("mac-cut-function", Fmac_cut_function, Smac_cut_function, 1, 2, 0,
    "Put the value of the string parameter to the Mac clipboard.")
  (value, push)
    Lisp_Object value, push;
{
  char *buf;
  int len, i;

  /* fixme: ignore the push flag for now */

  CHECK_STRING (value, 0);
  
  len = XSTRING (value)->size;
  buf = (char *) alloca (len);
  bcopy(XSTRING (value)->data, buf, len);
  
  /* convert to Mac-style eol's before sending to clipboard */
  for (i = 0; i < len; i++)
    if (buf[i] == '\n')
      buf[i] = '\r';

#if defined(TARGET_API_MAC_CARBON)
  {
      ScrapRef scrap;
      ClearCurrentScrap();
      if (GetCurrentScrap(&scrap) != noErr)
          return Qnil;
      if (PutScrapFlavor(scrap, kScrapFlavorTypeText,
                         0 /*ScrapFlavorFlags*/,len, buf) != noErr)
          return Qnil;
  }
#else
  ZeroScrap ();
  PutScrap (len, 'TEXT', buf);
#endif
  
  return Qnil;
}


static ComponentInstance gScriptingComponent;  /* an instance of the AppleScript component */
static OSAID gScriptContext;  /* the single script context used for all script executions */

/* Compile and execute the AppleScript SCRIPT and return the error
   status as function value.  A zero is returned if compilation and
   execution is successful, in which case RESULT returns a pointer to
   a string containing the resulting script value.  Otherwise, the Mac
   error code is returned and RESULT returns a pointer to an error
   string.  In both cases the caller should deallocate the storage
   used by the string pointed to by RESULT if it is non-NULL.  For
   documentation on the MacOS scripting architecture, see Inside
   Macintosh - Interapplication Communications: Scripting Components.  */
static long
DoAppleScript (char *script, char **result)
{
  AEDesc scriptDesc, resultDesc, errorDesc;
  OSErr error;
  OSAError osaerror;
  long length;

  *result = 0;

  error = AECreateDesc (typeChar, script, strlen(script), &scriptDesc);
  if (error)
    return error;

  osaerror = OSADoScript (gScriptingComponent, &scriptDesc, kOSANullScript, typeChar, kOSAModeNull, &resultDesc);

  if (osaerror == errOSAScriptError)  /* error executing AppleScript: retrieve error message */
    {
      if (!OSAScriptError (gScriptingComponent, kOSAErrorMessage, typeChar, &errorDesc))
        {
#if defined(TARGET_API_MAC_CARBON)
	  length = AEGetDescDataSize(&errorDesc);
#else
          HLock (errorDesc.dataHandle);
          length = GetHandleSize(errorDesc.dataHandle);
#endif
          *result = (char *) xmalloc (length + 1);
          if (*result)
            {
#if defined(TARGET_API_MAC_CARBON)
              AEGetDescData(&errorDesc,*result,length);
#else
              memcpy (*result, *(errorDesc.dataHandle), length);
#endif
              *(*result + length) = '\0';
            }
#if !defined(TARGET_API_MAC_CARBON)
          HUnlock (errorDesc.dataHandle);
#endif
          AEDisposeDesc (&errorDesc);
        }
    }
  else if (osaerror == noErr)  /* success: retrieve resulting script value */
    {
#if defined(TARGET_API_MAC_CARBON)
      length = AEGetDescDataSize(&resultDesc);
#else
      HLock (resultDesc.dataHandle);
      length = GetHandleSize(resultDesc.dataHandle);
#endif
      *result = (char *) xmalloc (length + 1);
      if (*result)
        {
#if defined(TARGET_API_MAC_CARBON)
	  AEGetDescData(&resultDesc,*result,length);
#else
          memcpy (*result, *(resultDesc.dataHandle), length);
#endif
          *(*result + length) = '\0';
        }
#if !defined(TARGET_API_MAC_CARBON)
      HUnlock (resultDesc.dataHandle);
#endif
    }

  AEDisposeDesc (&scriptDesc);
  AEDisposeDesc (&resultDesc);    

  return osaerror;
}

static void InitializeAppleScript()
{
  AEDesc nullDesc;
  OSAError osaerror;
  
  /* if open fails, gScriptingComponent is set to NULL.  Its subsequent use in
     OSA calls will fail with badComponentInstance error.  */
  gScriptingComponent = OpenDefaultComponent (kOSAComponentType, kAppleScriptSubtype);

  nullDesc.descriptorType = typeNull;
  nullDesc.dataHandle = 0;
  osaerror = OSAMakeContext (gScriptingComponent, &nullDesc, kOSANullScript, &gScriptContext);
  if (osaerror)
    gScriptContext = kOSANullScript;  /* use default context if create fails */
}

static void TerminateAppleScript()
{
  OSADispose (gScriptingComponent, gScriptContext);
  CloseComponent (gScriptingComponent);
}

DEFUN ("do-applescript", Fdo_applescript, Sdo_applescript, 1, 1, 0,
    "Compile and execute AppleScript SCRIPT and retrieve and return the result.\n\
If compilation and execution are successful, the resulting script\n\
value is returned as a string.  Otherwise the function aborts and\n\
displays the error message returned by the AppleScript scripting\n\
component.")
  (script)
    Lisp_Object script;
{
  char *result, *temp;
  Lisp_Object lisp_result;
  long status;

  CHECK_STRING (script, 0);
  
  status = DoAppleScript(XSTRING (script)->data, &result);
  if (status)
    {
      if (!result)
        {
          error ("Mac OS error %ld occurred while executing AppleScript", status);
        }
      else
        {
          /* Unfortunately only OSADoScript in DoAppleScript knows how
             how large the resulting script value or error message is
             going to be and therefore as caller memory must be
             deallocated here.  It is necessary to free the error
             message before calling error to avoid a memory leak.  */
          temp = (char *) alloca (strlen (result) + 1);
          strcpy (temp, result);
          xfree (result);
          error (temp);
        }
    }
  else
    {
      lisp_result = build_string (result);
      xfree (result);
      return lisp_result;
    }
}

#if !defined(TARGET_API_MAC_CARBON)
DEFUN ("mac-filename-to-unix", Fmac_filename_to_unix, Smac_filename_to_unix, 1, 1, 0,
    "Convert Macintosh filename to Unix form.")
  (mac_filename)
    Lisp_Object mac_filename;
{
  char unix_filename[MAXPATHLEN+1];

  CHECK_STRING (mac_filename, 0);
  
  if (Mac2UnixPathname(XSTRING (mac_filename)->data, unix_filename, MAXPATHLEN))
    return build_string (unix_filename);
  else
    return Qnil;
}

DEFUN ("unix-filename-to-mac", Funix_filename_to_mac, Sunix_filename_to_mac, 1, 1, 0,
    "Convert Unix filename to Mac form.")
  (unix_filename)
    Lisp_Object unix_filename;
{
  char mac_filename[MAXPATHLEN+1];

  CHECK_STRING (unix_filename, 0);
  
  if (Unix2MacPathname(XSTRING (unix_filename)->data, mac_filename, MAXPATHLEN))
    return build_string (mac_filename);
  else
    return Qnil;
}
#endif

/* Here are the set of functions assigned to the hooks defined in
   fontset.h to implement font sets.  */

/* Return a pointer to struct font_info of font FONT_IDX of frame F.  */
struct font_info *
mac_get_font_info (f, font_idx)
     FRAME_PTR f;
     int font_idx;
{
  return (FRAME_MAC_FONT_TABLE (f) + font_idx);
}

/* compare two strings ignoring case */
int stricmp (char *s, char *t)
{
  for ( ; tolower (*s) == tolower (*t); s++, t++)
    if (*s == '\0')
      return 0;
  return tolower (*s) - tolower (*t);
}

/* compare two strings ignoring case and handling wildcard */
int
wildstrieq (char *s1, char *s2)
{
  if (strcmp (s1, "*") == 0 || strcmp (s2, "*") == 0)
    return true;

  return stricmp (s1, s2) == 0;
}

/* Two font specs are considered to match if their foundry, family,
   weight, slant, and charset match.  */
int 
mac_font_match (char *mf, char *xf)
{
  char mFoundry[50], mFamily[50], mWeight[20], mSlant[2], mCharset[20];
  char xFoundry[50], xFamily[50], xWeight[20], xSlant[2], xCharset[20];

  if (sscanf (mf, "-%49[^-]-%49[^-]-%19[^-]-%1[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%19s",
              mFoundry, mFamily, mWeight, mSlant, mCharset) != 5)
    return 0;

  if (sscanf (xf, "-%49[^-]-%49[^-]-%19[^-]-%1[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%19s",
              xFoundry, xFamily, xWeight, xSlant, xCharset) != 5)
    return 0;

  return wildstrieq (mFoundry, xFoundry) && wildstrieq (mFamily, xFamily) &&
         wildstrieq (mWeight, xWeight) && wildstrieq (mSlant, xSlant) &&
         wildstrieq (mCharset, xCharset);
}

/* Convert a Mac font name/font number pair to an X font spec.  Notice
   that the charset field is deduced from the script number.  */
static char *
MacFontName2XFontName (char *mf, SInt16 fontNum)
{
  char foundry[32], family[32], cs[32];
  char xf[255], *result;

  if (sscanf (mf, "%31[^-]-%31[^-]-%31s", foundry, family, cs) != 3)
    {
      strcpy(foundry, "*");
      strcpy(family, mf);

      switch (FontToScript (fontNum))
      {
      case smTradChinese:
        strcpy(cs, "big5");
        break;
      case smSimpChinese:
        strcpy(cs, "gb2312");
        break;
      case smJapanese:
        strcpy(cs, "jisx0208-sjis");
        break;
      case smKorean:
        strcpy(cs, "ksc5601");
        break;        
      default:
        strcpy(cs, "mac-roman");
        break;
      }
    }

  sprintf(xf, "-%s-%s-*-*-*-*-*-*-*-*-*-*-%s",
          foundry, family, cs);
  
  result = (char *) xmalloc (strlen (xf) + 1);
  strcpy (result, xf);
  return result;
}

/* Convert an X font spec to the corresponding mac font name, which
   can then be passed to GetFNum after conversion to a Pascal string.
   For ordinary Mac fonts, this should just be their names, like
   "monaco", "Taipei", etc.  Fonts converted from the GNU intlfonts
   collection contain their charset designation in their names, like
   "ETL-Fixed-iso8859-1", "ETL-Fixed-koi8-r", etc.  Both types of font
   names are handled accordingly.  */
void
XFontName2MacFontName (char *xf, char *mf)
{
  char foundry[32], family[32], weight[20], slant[2], cs[32];

  strcpy (mf, "");

  if (sscanf (xf, "-%31[^-]-%31[^-]-%19[^-]-%1[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%31s",
              foundry, family, weight, slant, cs) != 5)
    return;

  if (strcmp (cs, "big5") == 0 || strcmp (cs, "gb2312") == 0 || strcmp (cs, "jisx0208-sjis") == 0 ||
      strcmp (cs, "ksc5601") == 0 || strcmp (cs, "mac-roman") == 0)
    strcpy(mf, family);
  else
    sprintf(mf, "%s-%s-%s", foundry, family, cs);
}

/* Sets up the table gFontNameTable to contain the list of all
   monospace fonts in the system the first time the table is used so
   that the Resource Manager need not be accessed every time this
   information is needed.  */
static void
InitFontNameTable ()
{
  GrafPtr port;
  int numMacFonts = CountResources('FOND');
  int i;
  Handle fontHandle;
  short id;
  ResType type;
  Str32 name;
  SInt16 fontNum, oldFontNum;

  GetPort(&port);  /* save the current font number used */
  oldFontNum = GetPortTextFont(port);

  for (i = 1; i <= numMacFonts; i++)  /* standard loop to get all available fonts */
    {
      fontHandle = GetIndResource ('FOND', i);
      if (fontHandle)
        {
          GetResInfo (fontHandle, &id, &type, name);
          GetFNum (name, &fontNum);
          if (fontNum != 0)
            {
              TextFont (fontNum);
	      /* Assuming that all non-Roman fonts are monospace
	         allows fonts to be found for some scripts which will
	         otherwise have no usable fonts (Korean, e.g.).  */
              if (FontToScript (fontNum) != smRoman || CharWidth ('m') == CharWidth ('i'))
                {
                  if (gFontNameTableSize == 0)
                    {
                      gFontNameTableSize = 16;
                      gFontNameTable = (char **) xmalloc (gFontNameTableSize * sizeof (char *));
                    }
                  else if (gFontNameCount >= gFontNameTableSize)
                    {
                      gFontNameTableSize += 16;
                      gFontNameTable = (char **) xrealloc (gFontNameTable, gFontNameTableSize * sizeof (char *));
                    }
                  p2cstr (name);
                  gFontNameTable[gFontNameCount++] = MacFontName2XFontName (name, fontNum);
                }
            }
          ReleaseResource (fontHandle);
        }
    }
    
  TextFont (oldFontNum);
}

/* After matching a font, mac_list_fonts calls MostSpecificName to
   return the most specific name (replacing wildcard specifiers with
   actual values) so that this can be returned to the caller.  This
   step is necessary because unlike X font specs, Mac font names do
   not contain weight, slant, font size, and charset information.
   mac_load_font maintains a list of font specs of fonts that have
   already been loaded to determine if a font (of a particular weight,
   size, etc. variant) needs to be loaded.  If less specific names are
   used, the use of a bold variant of a font will not cause a new
   entry to be properly created for it, for example.  */
void 
MostSpecificName (char *msn, char *mf, char *xf)
{
  char mFoundry[50], mFamily[50], mWeight[20], mSlant[2], mCharset[20];
  char xFoundry[50], xFamily[50], xWeight[20], xSize[10], xSlant[2], xCharset[20];

  if (sscanf (mf, "-%49[^-]-%49[^-]-%19[^-]-%1[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%19s",
              mFoundry, mFamily, mWeight, mSlant, mCharset) != 5)
    abort ();

  if (sscanf (xf, "-%49[^-]-%49[^-]-%19[^-]-%1[^-]-%*[^-]-%*[^-]-%9[^-]-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%19s",
              xFoundry, xFamily, xWeight, xSlant, xSize, xCharset) != 6)
    abort ();

  sprintf(msn, "-%s-%s-%s-%s-*-*-%s-*-*-*-*-*-%s",
          strcmp (mFoundry, "*") != 0 ? mFoundry : xFoundry,
          strcmp (mFamily, "*") != 0 ? mFamily : xFamily,
          strcmp (mWeight, "*") != 0 ? mWeight : xWeight,
          strcmp (mSlant, "*") != 0 ? mSlant : xSlant,
          xSize,
          strcmp (mCharset, "*") != 0 ? mCharset : xCharset);
}

/* Return a list of at most MAXNAMES font specs matching the one in
   PATTERN.  Note that each '*' in the PATTERN matches exactly one
   field of the font spec, unlike X in which an '*' in a font spec can
   match a number of fields.  The result is in the Mac implementation
   all fonts must be specified by a font spec with all 13 fields
   (although many of these can be "*'s").  */
Lisp_Object
mac_list_fonts (struct frame *f,
                Lisp_Object pattern,
                int size,
                int maxnames)
{
  char *ptnstr;
  Lisp_Object newlist = Qnil;
  int n_fonts = 0;
  int i;
  char fontName[256];

  if (gFontNameTable == NULL)  /* initialize when first used */
    InitFontNameTable ();
  
  ptnstr = XSTRING (pattern)->data;

  for (i = 0; i < gFontNameCount; i++)  /* scan through the and return matching ones */
    {
      if (mac_font_match (gFontNameTable[i], ptnstr))
        {
          MostSpecificName (fontName, gFontNameTable[i], ptnstr);
          newlist = Fcons (build_string (fontName), newlist);
          n_fonts++;
          if (n_fonts >= maxnames)
            break;
        }
    }

  return newlist;
}

/* MacLoadQueryFont creates and returns an internal representation for
   a font in a MacFontStruct struct (similar in function to
   XLoadQueryFont in X).  There is really no concept corresponding to
   "loading" a font on the Mac.  But we check its existence and find
   the font number and all other information for it and store them in
   the returned MacFontStruct.  */
MacFontStruct *
MacLoadQueryFont (char *fontname)
{
  GrafPtr port;
  SInt16 oldFontNum, oldFontSize;
  Style oldFontFace;
  Str32 mfontname;
  SInt16 fontnum;
  int size;
  Style fontface = normal;
  MacFontStruct *font;
  FontInfo theFontInfo;
  char sWeight[7], cSlant;

  GetPort(&port);  /* save the current font number used */
  oldFontNum = GetPortTextFont(port); //port->txFont;
  oldFontSize = GetPortTextSize(port); //port->txSize;
  oldFontFace = GetPortTextFace(port); //port->txFace;

  if (sscanf (fontname, "-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%*[^-]-%d-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%*s", &size) != 1)
    size = kDefaultFontSize;

  if (sscanf (fontname, "-%*[^-]-%*[^-]-%6[^-]-%*c-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%*s", sWeight) == 1)
    if (strcmp (sWeight, "bold") == 0)
      fontface |= bold;

  if (sscanf (fontname, "-%*[^-]-%*[^-]-%*[^-]-%c-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*[^-]-%*c-%*[^-]-%*s", &cSlant) == 1)
    if (cSlant == 'i')
      fontface |= italic;

  XFontName2MacFontName (fontname, mfontname);
  c2pstr (mfontname);
  GetFNum (mfontname, &fontnum);
  if (fontnum == 0)
    return NULL;
    
  font = (MacFontStruct *) xmalloc (sizeof (struct MacFontStruct));
  
  font->mFontNum = fontnum;
  font->mFontSize = size;
  font->mFontFace = fontface;
  font->mScriptCode = FontToScript (fontnum);
  font->mTwoByte = font->mScriptCode == smJapanese || font->mScriptCode == smTradChinese ||
                   font->mScriptCode == smSimpChinese || font->mScriptCode == smKorean;

  TextFont (fontnum);
  TextSize (size);
  TextFace (fontface);
  
  GetFontInfo (&theFontInfo);
  font->mHeight = theFontInfo.ascent + theFontInfo.descent + theFontInfo.leading;
  /* mwp->mCharWidth = theFontInfo.widMax; Incorrectly returns 15 for
     12-point Monaco! */
  font->mWidth = CharWidth ('m');
  font->mAscent = theFontInfo.ascent;
  font->mDescent = theFontInfo.descent;
  font->mLeading = theFontInfo.leading;

  TextFont (oldFontNum);  /* restore previous font number and size */
  TextSize (oldFontSize);
  TextFace (oldFontFace);
  
  return font;
}

/* Load font named FONTNAME of the size SIZE for frame F, and return a
   pointer to the structure font_info while allocating it dynamically.
   If SIZE is 0, load any size of font.  If loading is failed, return
   NULL.  */   
struct font_info *
mac_load_font (struct frame *f, char *fontname, int size)
{
  Lisp_Object font_names;
  struct font_info *fontp;
  MacFontStruct *font;
  int i;
  Lisp_Object tail;

  if (!fontname)
    return NULL;

  font_names = mac_list_fonts (f, build_string (fontname), size, 1);
  if (NILP (font_names))
    return NULL;

  /* if the font has been loaded, return it */
  for (i = 0; i < n_fonts; i++)
    for (tail = font_names; CONSP (tail); tail = XCONS (tail)->cdr)
      if (!strcmp (font_table[i].name, XSTRING (XCONS (tail)->car)->data))
        return font_table + i;

  /* Do we need to create the table?  */
  if (font_table_size == 0)
    {
      font_table_size = 16;
      font_table = (struct font_info *) xmalloc (font_table_size * sizeof (struct font_info));
    }
  /* Do we need to grow the table?  */
  else if (n_fonts >= font_table_size)
    {
      font_table_size *= 2;
      font_table = (struct font_info *) xrealloc (font_table, font_table_size * sizeof (struct font_info));
    }

  fontp = font_table + n_fonts;

  font = MacLoadQueryFont (XSTRING (XCONS (font_names)->car)->data);
  if (!font)
    return NULL;
  
  /* Now fill in the slots of *FONTP.  */
  fontp->font = (void *) font;
  fontp->font_idx = n_fonts;
  fontp->name = (char *) xmalloc (strlen (fontname) + 1);
  bcopy (fontname, fontp->name, strlen (fontname) + 1);

  fontp->full_name = fontp->name;

  fontp->size = font->mFontSize;
  fontp->height = font->mHeight;

  if (font->mScriptCode == smJapanese)
    fontp->encoding[1] = 4;
  else
      fontp->encoding[1] = FONT_ENCODING_NOT_DECIDED;

  /* The following three values are set to 0 under W32, which is
     what they get set to if XGetFontProperty fails under X.  */
  fontp->baseline_offset = 0;
  fontp->relative_compose = 0;
  fontp->default_ascent = 0;

  n_fonts++;

  return fontp;
}

void
mac_check_window_system ()
{
  /* just return: if you call this routine, you are running on a Mac */
}

/* Find a CCL program for a font specified by FONTP, and set the
   member `encoder' of the structure.  Taken from xterm.c.  */
void
mac_find_ccl_program (fontp)
     struct font_info *fontp;
{
  extern Lisp_Object Vfont_ccl_encoder_alist, Vccl_program_table;
  extern Lisp_Object Qccl_program_idx;
  extern Lisp_Object resolve_symbol_ccl_program ();
  Lisp_Object list, elt, ccl_prog, ccl_id;

  for (list = Vfont_ccl_encoder_alist; CONSP (list); list = XCONS (list)->cdr)
    {
      elt = XCONS (list)->car;
      if (CONSP (elt)
	  && STRINGP (XCONS (elt)->car)
	  && (fast_c_string_match_ignore_case (XCONS (elt)->car, fontp->name)
	      >= 0))
	{
	  if (SYMBOLP (XCONS (elt)->cdr) &&
	      (!NILP (ccl_id = Fget (XCONS (elt)->cdr, Qccl_program_idx))))
	    {
	      ccl_prog = XVECTOR (Vccl_program_table)->contents[XUINT (ccl_id)];
	      if (!CONSP (ccl_prog)) continue;
	      ccl_prog = XCONS (ccl_prog)->cdr;
	    }
	  else
	    {
	      ccl_prog = XCONS (elt)->cdr;
	      if (!VECTORP (ccl_prog)) continue;
	    }
	    
	  fontp->font_encoder
	    = (struct ccl_program *) xmalloc (sizeof (struct ccl_program));
	  setup_ccl_program (fontp->font_encoder,
			     resolve_symbol_ccl_program (ccl_prog));
	  break;
	}
    }
}

/* Return a pointer to struct font_info of a font named FONTNAME for
   frame F.  If no such font is loaded, return NULL.  */
struct font_info*
mac_query_font (struct frame *f, char *fontname)
{
  int i;
  struct font_info *pfi;

  pfi = FRAME_MAC_FONT_TABLE (f);

  for (i = 0; i < n_fonts ;i++, pfi++)
    if (strcmp (pfi->name, fontname) == 0)
      return pfi;

  return NULL;
}

/* Give frame F the font named FONTNAME as its default font, and
   return the full name of that font.  FONTNAME may be a wildcard
   pattern; in that case, we choose some font that fits the pattern.
   The return value shows which font we chose.  */
Lisp_Object
mac_new_font (f, fontname)
     struct frame *f;
     register char *fontname;
{
  mac_output *mwp;
  struct font_info *fontp
    = fs_load_font (f, FRAME_MAC_FONT_TABLE (f), CHARSET_ASCII, fontname, -1);

  if (!fontp)
    return Qnil;

  f->output_data.mac->font = (MacFontStruct *) (fontp->font);
  f->output_data.mac->fontset = -1;
  f->output_data.mac->font_baseline = f->output_data.mac->font->mAscent + fontp->baseline_offset;
  
  frame_update_line_height (f);
  x_set_window_size (f, 0, f->width, f->height);

  return build_string (fontp->full_name);
}

/* Give frame F the fontset named FONTSETNAME as its default font, and
   return the full name of that fontset.  FONTSETNAME may be a
   wildcard pattern; in that case, we choose some fontset that fits
   the pattern.  The return value shows which fontset we chose.  */
Lisp_Object
mac_new_fontset (f, fontsetname)
     struct frame *f;
     char *fontsetname;
{
  int fontset = fs_query_fontset (f, fontsetname);
  struct fontset_info *fontsetp;
  Lisp_Object result;

  if (fontset < 0)
    return Qnil;

  if (FRAME_FONTSET (f) == fontset)
    /* This fontset is already set in frame F.  There's nothing more
       to do.  */
    return build_string (fontsetname);

  fontsetp = FRAME_FONTSET_DATA (f)->fontset_table[fontset];

  if (!fontsetp->fontname[CHARSET_ASCII])
    /* This fontset doesn't contain ASCII font.  */
    return Qnil;

  result = mac_new_font (f, fontsetp->fontname[CHARSET_ASCII]);

  if (!STRINGP (result))
    /* Can't load ASCII font.  */
    return Qnil;

  /* Since mac_new_font doesn't update any fontset information, do it now.  */
  FRAME_FONTSET(f) = fontset;
  FS_LOAD_FONT (f, FRAME_MAC_FONT_TABLE (f),
		CHARSET_ASCII, XSTRING (result)->data, fontset);

  return build_string (fontsetname);
}

void
mac_set_frame_fontset (f, arg, oldVal)
     struct frame *f;
     Lisp_Object arg;
     Lisp_Object oldVal;
{
  Lisp_Object result;
  Lisp_Object fontset_name;
  Lisp_Object frame;

  CHECK_STRING (arg, 1);

  fontset_name = Fquery_fontset (arg, Qnil);

  BLOCK_INPUT;
  result = (STRINGP (fontset_name)
            ? mac_new_fontset (f, XSTRING (fontset_name)->data)
            : mac_new_font (f, XSTRING (arg)->data));
  UNBLOCK_INPUT;
  
  if (EQ (result, Qnil))
    error ("Font \"%s\" is not defined", XSTRING (arg)->data);
  else if (EQ (result, Qt))
    error ("The characters of the given font have varying widths");
  else if (STRINGP (result))
    {
      recompute_basic_faces (f);
      store_frame_param (f, Qfont, result);
    }
  else
    abort ();

  XSETFRAME (frame, f);
  call1 (Qface_set_after_frame_default, frame);
}

#include "x-list-font.c"

MacFontStruct *
load_font (f, name)
     struct frame *f;
     Lisp_Object name;
{
  MacFontStruct *font;

  if (NILP (name))
    return (MacFontStruct *) FACE_DEFAULT;

  CHECK_STRING (name, 0);
  BLOCK_INPUT;
  font = MacLoadQueryFont ((char *) XSTRING (name)->data);
  UNBLOCK_INPUT;

  if (! font)
    Fsignal (Qerror, Fcons (build_string ("undefined font"),
			    Fcons (name, Qnil)));
  return font;
}

void
unload_font (f, font)
     struct frame *f;
     MacFontStruct *font;
{
  int len = FRAME_N_COMPUTED_FACES (f);
  int i;

  if (font == NULL || font == ((MacFontStruct *) FACE_DEFAULT))
    return;

  BLOCK_INPUT;
  /* Invalidate any computed faces which use this font,
     and free their GC's if they have any.  */
  for (i = 2; i < len; i++)
    {
      struct face *face = FRAME_COMPUTED_FACES (f)[i];
      if (face->font == font)
	{
	  face->gc = 0;
	  /* This marks the computed face as available to reuse.  */
	  face->font = 0;
	}
    }

  xfree (font);
  UNBLOCK_INPUT;
}

/* Update the line_height of frame F according to the biggest font in
   any face.  Return nonzero if if line_height changes.  */
int
frame_update_line_height (f)
     FRAME_PTR f;
{
  int i;
  int fontset = f->output_data.mac->fontset;
  int biggest = (fontset > 0
		 ? FRAME_FONTSET_DATA (f)->fontset_table[fontset]->height
		 : FONT_HEIGHT (f->output_data.mac->font));

  for (i = 0; i < f->output_data.mac->n_param_faces; i++)
    if (f->output_data.mac->param_faces[i] != 0
	&& f->output_data.mac->param_faces[i]->font != (XFontStruct *) FACE_DEFAULT)
      {
	int height = ((fontset = f->output_data.mac->param_faces[i]->fontset) > 0
		      ? FRAME_FONTSET_DATA (f)->fontset_table[fontset]->height
		      : FONT_HEIGHT (f->output_data.mac->param_faces[i]->font));

	if (height > biggest)
	  biggest = height;
      }

  if (biggest == f->output_data.mac->mLineHeight)
    return 0;

  f->output_data.mac->mLineHeight = biggest;
  return 1;
}

/* Intialize AppleEvent dispatcher table for the required events.  */
void
InitRequiredAppleEvents ()
{
  OSErr err;
  long result;

  /* Make sure we have apple events before starting.  */
  err = Gestalt (gestaltAppleEventsAttr, &result);
  if (err != noErr)
    abort ();

  if (!(result & (1 << gestaltAppleEventsPresent)))
    abort ();
  
  err = AEInstallEventHandler(kCoreEventClass, kAEOpenApplication,
                              NewAEEventHandlerUPP ((AEEventHandlerProcPtr) doAEOpenApplication),
                              0L, false);
  if (err != noErr)
    abort ();
    
  err = AEInstallEventHandler(kCoreEventClass, kAEOpenDocuments,
                              NewAEEventHandlerUPP ((AEEventHandlerProcPtr) doAEOpenDocuments),
                              0L, false);
  if (err != noErr)
    abort ();

  err = AEInstallEventHandler(kCoreEventClass, kAEPrintDocuments,
                              NewAEEventHandlerUPP ((AEEventHandlerProcPtr) doAEPrintDocuments),
                              0L, false);
  if (err != noErr)
    abort ();

  err = AEInstallEventHandler(kCoreEventClass, kAEQuitApplication,
                              NewAEEventHandlerUPP ((AEEventHandlerProcPtr) doAEQuitApplication),
                              0L, false);
  if (err != noErr)
    abort ();
}

/* Open Application Apple Event */
static pascal OSErr doAEOpenApplication(const AppleEvent *pAE, AppleEvent *pReply, long pRefcon)
{
  return noErr;
}

/* Defined in sysdep.c.  */
extern int
PathFromVolDirName (char *path, int maxPathLen, short vRefNum, long dirID, ConstStr255Param name);

/* Called when we receive an AppleEvent with an ID of
   "kAEOpenDocuments".  This routine gets the direct parameter,
   extracts the FSSpecs in it, and puts their names on a list.  */
static pascal OSErr     
doAEOpenDocuments(AppleEvent* message, AppleEvent *pReply, long pRefcon)
{
  OSErr err, err2;
  AEDesc theDesc;
  AEKeyword keyword;
  DescType actualType;
  Size actualSize;

  err = AEGetParamDesc (message, keyDirectObject, typeAEList, &theDesc);
  if (err != noErr)
    goto __errDesc;

  /* Check to see that we got all of the required parameters from the
     event descriptor.  For an 'odoc' event this should just be the
     file list.  */
  err = AEGetAttributePtr(message, keyMissedKeywordAttr, typeWildCard,  &actualType, (Ptr) &keyword,
                          sizeof (keyword), &actualSize);
  /* No error means that we found some unused parameters.
     errAEDescNotFound means that there are no more parameters.  If we
     get an error code other than that, flag it.  */
  if ((err == noErr) || (err != errAEDescNotFound))
    {
      err = errAEEventNotHandled;
      goto __errExit;
    }
  err = noErr;

  /* Got all the parameters we need.  Now, go through the direct
     object list and parse it up.  */
  {
    long numFilesToOpen;

    err = AECountItems (&theDesc, &numFilesToOpen);
    if (err == noErr)
      {
        int i;
        
        /* AE file list is one based so just use that for indexing here.  */
        for (i = 1; (err == noErr) && (i <= numFilesToOpen); i++) {
          FSSpec fs;
	  Str255 pathName, unixPathName;

          err = AEGetNthPtr(&theDesc, i, typeFSS, &keyword, &actualType, (Ptr) &fs, sizeof (fs),
                            &actualSize);
          if (err != noErr) break;

	  if (PathFromVolDirName (pathName, 255, fs.vRefNum, fs.parID, fs.name) &&
	      Mac2UnixPathname (pathName, unixPathName, 255))
            gDragAndDropFileList = Fcons (build_string (unixPathName), gDragAndDropFileList);
        }
      }
  }

__errExit:
  /* Nuke the coerced file list in any case */
  err2 = AEDisposeDesc(&theDesc);

__errDesc:
  /* InvalRect(&(gFrontMacWindowP->mWP->portRect)); */
  return err;
}

/* Print Document Apple Event */
static pascal OSErr
doAEPrintDocuments (const AppleEvent *pAE, AppleEvent *pReply, long pRefcon)
{
  return errAEEventNotHandled;
}

static pascal OSErr
doAEQuitApplication (AppleEvent* message, AppleEvent *pReply, long pRefcon)
{
  /* FixMe: Do we need an unwind-protect or something here?  And what
     do we do about unsaved files. Currently just forces quit rather
     than doing recursive callback to get user input.  */

  gDone = true;

  /* Fkill_emacs doesn't return.  We have to return. (TI) */
  return noErr;
}


/* When converting from Mac to Unix pathnames, /'s in folder names
   are converted to :'s.  This function, used in copying folder names,
   performs a strncat and converts all character a to b in the
   copy of the string s2 appended to the end of s1.  */
void StringCatNReplace(char *s1, const char *s2, int n, char a, char b)
{
  int l1 = strlen(s1);
  int l2 = strlen(s2);
  char *p = s1 + l1;
  int i;
  
  strncat(s1, s2, n);
  for (i = 0; i < l2; i++)
    {
      if (*p == a)
        *p = b;
      p++;
    }
}

/* Convert a Mac pathname to Unix form.  A Mac full pathname is one
   that does not begin with a ':' and contains at least one ':'. A Mac
   full pathname causes an '/' to be prepended to the Unix pathname.
   The algorithm for the rest of the pathname is as follows:
     For each segment between two ':',
       if it is non-null, copy as is and then add a '/' at the end,
       otherwise, insert a "../" into the Unix pathname.
   Returns 1 if successful; 0 if fails.  */
   
int
Mac2UnixPathname (const char *mfn, char *ufn, int ufnbuflen)
{
  const char *p, *q, *pe;
	
  strcpy (ufn, "");
	
  if (*mfn == '\0')
    return 1;
	
  p = strchr (mfn, ':');
  if (p != 0 && p != mfn)  /* full pathname */
    strcat (ufn, "/");
		
  p = mfn;
  if (*p == ':')
    p++;

  pe = mfn + strlen (mfn);
  while (p < pe)
    {
      q = strchr (p, ':');
      if (q)
	{
	  if (q == p)  /* two consecutive ':' */
	    {
	      if (strlen (ufn) + 3 >= ufnbuflen)
		return 0;
	      strcat (ufn, "../");
	    }
	  else
	    {
	      if (strlen (ufn) + (q - p) + 1 >= ufnbuflen)
		return 0;
	      StringCatNReplace (ufn, p, q - p, '/', ':');
	      strcat (ufn, "/");
	    }
	  p = q + 1;
	}
      else
	{
	  if (strlen (ufn) + (pe - p) >= ufnbuflen)
	    return 0;
	  StringCatNReplace (ufn, p, pe - p, '/', ':');  /* no separator for last one */
	  p = pe;
	}
    }
	
  return 1;
}

/* Determine the path name of the file specified by VREFNUM, DIRID, and NAME and place
   that in the buffer PATH of length MAXPATHLEN.  */
int
PathFromVolDirName (char *path, int maxPathLen, short vRefNum, long dirID, ConstStr255Param name)
{
  Str255 dirName;
  CInfoPBRec cipb;
  OSErr err;

  if (strlen (name) > maxPathLen)
    return 0;

  memcpy (dirName, name, name[0]+1);
  memcpy (path, name, name[0]+1);
  p2cstr (path);

  cipb.dirInfo.ioDrParID = dirID;
  cipb.dirInfo.ioNamePtr = dirName;

  do
    {
      cipb.dirInfo.ioVRefNum = vRefNum;
      cipb.dirInfo.ioFDirIndex = -1;
      cipb.dirInfo.ioDrDirID = cipb.dirInfo.ioDrParID;  /* go up to parent each time */

      err = PBGetCatInfo (&cipb, false);
      if (err != noErr)
        return 0;
      
      p2cstr (dirName);
      if (strlen (dirName) + strlen (path) + 1 >= maxPathLen)
        return 0;

      strcat (dirName, ":");
      strcat (dirName, path);  /* attach to front since going up directory tree */
      strcpy (path, dirName);
    }
  while (cipb.dirInfo.ioDrDirID != fsRtDirID);  /* until volume's root directory */
  
  return 1;  /* success */
}


void
syms_of_macterm ()
{
  get_font_info_func = mac_get_font_info;
  list_fonts_func = mac_list_fonts;
  load_font_func = mac_load_font;
  find_ccl_program_func = mac_find_ccl_program;
  query_font_func = mac_query_font;
  set_frame_fontset_func = mac_set_frame_fontset;
  check_window_system_func = mac_check_window_system;

  syms_of_fontset ();  /* should really set this in emacs.c */
  
  DEFVAR_LISP ("mac-command-key-is-meta", &Vmac_command_key_is_meta,
    "Non-nil means that the command key is used as the Emacs meta key.\n\
Otherwise the option key is used.");
  Vmac_command_key_is_meta = Qnil;

  DEFVAR_INT ("mac-keyboard-text-encoding", &mac_keyboard_text_encoding,
    "One of the Text Encoding Base constant values defined in the\n\
Basic Text Constants section of Inside Macintosh - Text Encoding\n\
Conversion Manager.  Its value determines the encoding characters\n\
typed at the Mac keyboard (presumed to be in the MacRoman encoding)\n\
will convert into.  E.g., if it is set to kTextEncodingMacRoman (0),\n\
its default value, no conversion takes place.  If it is set to\n\
kTextEncodingISOLatin1 (0x201) or kTextEncodingISOLatin2 (0x202),\n\
characters typed on Mac keyboard are first converted into the\n\
ISO Latin-1 or ISO Latin-2 encoding, respectively before being\n\
passed to Emacs.  Together with Emacs's set-keyboard-coding-system\n\
command, this enables the Mac keyboard to be used to enter non-ASCII\n\
characters directly.");
  mac_keyboard_text_encoding = kTextEncodingMacRoman;

  DEFVAR_LISP ("x-resource-name", &Vx_resource_name,
    "The name Emacs uses to look up resources; for internal use only.\n\
`x-get-resource' uses this as the first component of the instance name\n\
when requesting resource values.\n\
Emacs initially sets `x-resource-name' to the name under which Emacs\n\
was invoked, or to the value specified with the `-name' or `-rn'\n\
switches, if present.");
  Vx_resource_name = Qnil;

  Qleft = intern ("left");
  staticpro (&Qleft);
  Qtop = intern ("top");
  staticpro (&Qtop);

  Qface_set_after_frame_default = intern ("face-set-after-frame-default");
  staticpro (&Qface_set_after_frame_default);

  Qbackground_color = intern ("background-color");
  staticpro (&Qbackground_color);
  Qforeground_color = intern ("foreground-color");
  staticpro (&Qforeground_color);
  Qinternal_border_width = intern("internal-border-width");
  staticpro (&Qinternal_border_width);
  Qforeground_color = intern("foreground-color");
  staticpro (&Qforeground_color);
  Qbackground_color = intern("background-color");
  staticpro (&Qbackground_color);
  Qvertical_scroll_bars = intern("vertical-scroll-bars");
  staticpro (&Qvertical_scroll_bars);
  Qmouse_color = intern("mouse-color");
  staticpro (&Qmouse_color);
  Qcursor_color = intern("cursor-color");
  staticpro (&Qcursor_color);
  Qborder_color = intern("border-color");
  staticpro (&Qborder_color);
  Qmenu_bar_lines = intern("menu-bar-lines");
  staticpro (&Qmenu_bar_lines);
  Qscroll_bar_width = intern("scroll-bar-width");
  staticpro (&Qscroll_bar_width);
  Qbuffer_predicate = intern("buffer-predicate");
  staticpro (&Qbuffer_predicate);
  Qicon_type = intern("icon-type");
  staticpro (&Qicon_type);
  Qauto_raise = intern("auto-raise");
  staticpro (&Qauto_raise);
  Qauto_lower = intern("auto-lower");
  staticpro (&Qauto_lower);
  Qcursor_type = intern("cursor-type");
  staticpro (&Qcursor_type);
  Qbox = intern("box");
  staticpro (&Qbox);
  Qnone = intern("none");
  staticpro (&Qnone);
 
  defsubr (&Sx_create_frame);
  defsubr (&Sx_get_resource);
  defsubr (&Smac_paste_function);
  defsubr (&Smac_cut_function);
  defsubr (&Sdo_applescript);
#if !defined(TARGET_API_MAC_CARBON)
  defsubr (&Smac_filename_to_unix);
  defsubr (&Sunix_filename_to_mac);
#endif  
  defsubr (&Sx_list_fonts);
}
