/* Defines new terminal type for GNU Emacs on the Macintosh.
   Copyright (C) 1999 Free Software Foundation, Inc.
   Author: Andrew Choi <choi@cs.hku.hk>

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#ifndef _MACTERM_H_
#define _MACTERM_H_

#ifdef NON_OPAQUE
  #ifdef TARGET_API_MAC_CARBON
#if 0
    #undef DARWIN
    #undef macintosh
    #ifdef Z
      #undef Z
      #define __NEED_Z_DEFINE
    #endif
    #ifdef O_ASYNC
      #undef O_ASYNC
      #undef O_NDELAY
      #undef O_NONBLOCK
      #define __NEED_O_ASYNC
    #endif
    #undef min
#endif
    #include <Carbon/Carbon.h>
#if 0
    #define DARWIN
    #define macintosh
    #ifdef __NEED_Z_DEFINE
      #define Z (current_buffer->text->z)
    #endif
    #ifdef __NEED_O_ASYNC
      #define	O_ASYNC		0x0040		/* signal pgrp when data ready */
      #define	O_NDELAY	O_NONBLOCK	/* compat */
      #define	O_NONBLOCK	0x0004		/* no delay */
    #endif
#endif
  #else
    #include <Quickdraw.h>
    #include <MacTypes.h>
  #endif
#endif

#include "frame.h"

#if !defined(NON_OPAQUE)
 typedef unsigned char 					Style;
 typedef short SInt16;
#endif

struct MacFontStruct {
  SInt16 mFontNum;		/* font number of font used in this window */
  short mScriptCode;	/* Mac OS script code for font used */
  int mFontSize;		/* size of font */
  Style mFontFace;		/* plain, bold, italics, etc. */
  int mHeight;		/* height of one line of text in pixels */
  int mWidth;		/* width of one character in pixels */
  int mAscent;
  int mDescent;
  int mLeading;
  char mTwoByte;  /* true for two-byte font */
};
typedef struct MacFontStruct MacFontStruct;
typedef struct MacFontStruct XFontStruct;

typedef int GC;
typedef int Pixmap;
typedef int Display;

/* The collection of data describing a window on the Mac.  Functions
   defined in macterm.c */
struct mac_output {
#ifdef NON_OPAQUE
  WindowPtr mWP;		/* pointer to QuickDraw window */
#else
  void *mWP;
#endif
  FRAME_PTR mFP;		/* points back to the frame struct */
  int mNumCols;			/* number of characters per column */
  int mNumRows;			/* number of characters per row */
  int mLineHeight;		/* height of one line of text in pixels */
  int mCharWidth;		/* width of one character in pixels */
  int mHomeX;			/* X pixel coordinate of lower left corner of character at (0, 0) */
  int mHomeY;			/* Y pixel coordinate of lower left corner of character at (0, 0) */
  int mHighlight;		/* current highlight state (0 = off). */
  int mTermWinSize;		/* num of lines from top of window affected by ins_del_lines; set by set_terminal_window. */
  /* stuffs used by xfaces.c */
  XFontStruct *font;
  int fontset;
  struct face **param_faces;
  int n_param_faces;
  struct face **computed_faces;
  int n_computed_faces;
  int size_computed_faces;
  unsigned long background_pixel;
  unsigned long foreground_pixel;
  unsigned long cursor_pixel;
  /* stuffs needed after defining FRAME_WINDOW_P as FRAME_MAC_P */
  int left_pos;
  int top_pos;
  /* used by dumpglyphs */
  int font_baseline;
};

typedef struct mac_output mac_output;

/* A table of all the fonts we have already loaded.
   On the Mac there is one global font table.  */
extern struct font_info *font_table;

/* The current capacity of font_table.  */
extern int font_table_size;

/* The number of fonts actually stored in font_table.
   font_table[n] is used and valid iff 0 <= n < n_fonts.
   0 <= n_fonts <= font_table_size.  */
extern int n_fonts;

extern int unibyte_display_via_language_environment;

extern struct face *intern_face (struct frame *, struct face *);

/* Get at the computed faces of an X window frame.  */
#define FRAME_PARAM_FACES(f) ((f)->output_data.mac->param_faces)
#define FRAME_N_PARAM_FACES(f) ((f)->output_data.mac->n_param_faces)
#define FRAME_DEFAULT_PARAM_FACE(f) (FRAME_PARAM_FACES (f)[0])
#define FRAME_MODE_LINE_PARAM_FACE(f) (FRAME_PARAM_FACES (f)[1])

#define FRAME_COMPUTED_FACES(f) ((f)->output_data.mac->computed_faces)
#define FRAME_N_COMPUTED_FACES(f) ((f)->output_data.mac->n_computed_faces)
#define FRAME_SIZE_COMPUTED_FACES(f) ((f)->output_data.mac->size_computed_faces)
#define FRAME_DEFAULT_FACE(f) ((f)->output_data.mac->computed_faces[0])
#define FRAME_MODE_LINE_FACE(f) ((f)->output_data.mac->computed_faces[1])

#define FRAME_FONT(f) ((f)->output_data.mac->font)
#define FRAME_FONTSET(f) ((f)->output_data.mac->fontset)
#define FRAME_X_DISPLAY(f) ((Display *) 0)

#define FRAME_FOREGROUND_PIXEL(f) ((f)->output_data.mac->foreground_pixel)
#define FRAME_BACKGROUND_PIXEL(f) ((f)->output_data.mac->background_pixel)

#define FONT_WIDTH(f) ((f)->mWidth)
#define FONT_HEIGHT(f) ((f)->mHeight)

#define XFreeGC (void)
#define same_size_fonts(foo,bar) (1)
#define unload_color(p1,p2)
#define x_destroy_bitmap(p1,p2)
#define load_pixmap(p1,p2,p3,p4) (0)

/* This is the 'font_info *' which frame F has.  */
#define FRAME_MAC_FONT_TABLE(f) font_table

#endif
